"""Module for launching MAPDL locally."""
import platform
from glob import glob
import re
import warnings
import os
import appdirs
import tempfile
import socket

from pyansys.misc import is_float, random_string
from pyansys.errors import LockFileException, VersionError

# settings directory
SETTINGS_DIR = appdirs.user_data_dir('pyansys')
if not os.path.isdir(SETTINGS_DIR):
    try:
        os.makedirs(SETTINGS_DIR)
    except:
        warnings.warn('Unable to create settings directory.\n' +
                      'Will be unable to cache ANSYS executable location')

CONFIG_FILE = os.path.join(SETTINGS_DIR, 'config.txt')
ALLOWABLE_MODES = ['corba', 'console']

LOCALHOST = '127.0.0.1'
MAPDL_DEFAULT_PORT = 50052


def port_in_use(port, ip='localhost'):
    """Returns True when a port is in use at the given ip"""
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        return s.connect_ex((ip, port)) == 0


def find_ansys():
    """Searches for ansys path within environmental variables.

    Reutrns
    -------
    ansys_exe_path : str
        Full path to ANSYS executable

    version : float
        Version of ANSYS
    """
    ansys_sysdir_var = 'ANSYS_SYSDIR'
    paths = {}
    for var in os.environ:
        if 'ANSYS' in var and '_DIR' in var:
            # add path if valid
            path = os.environ[var]
            if os.path.isdir(path):

                # add path if version number is in path
                version_str = var[5:8]
                if is_float(version_str):
                    paths[int(version_str)] = path

    # On Linux ansys is usually installed at the root directory at
    # /usr/ansys_inc
    # /ansys_inc
    if not paths and os.name == 'posix':
        for directory in ['/usr/ansys_inc/', '/ansys_inc/']:
            if os.path.isdir(directory):
                # construct ansys path
                for subdirectory in glob('%s*/' % directory):
                    ver = os.path.basename(os.path.normpath('/usr/ansys_inc/v202/'))
                    groups = re.findall(r'v\d\d\d', ver)
                    if groups:
                        ver = int(groups[0][1:])
                        paths[ver] = os.path.join(subdirectory, 'ansys')

    if not paths:
        return '', ''

    # check through all available paths and return the latest version
    while paths:
        version = max(paths.keys())
        ansys_path = paths[version]

        if ansys_sysdir_var in os.environ:
            sysdir = os.environ[ansys_sysdir_var]
            ansys_bin_path = os.path.join(ansys_path, 'bin', sysdir)
            if 'win' in sysdir:
                ansys_exe = 'ansys%d.exe' % version
            else:
                ansys_exe = 'ansys%d' % version
        else:
            ansys_bin_path = os.path.join(ansys_path, 'bin')
            ansys_exe = 'ansys%d' % version

        ansys_exe_path = os.path.join(ansys_bin_path, ansys_exe)
        if os.path.isfile(ansys_exe_path):
            break
        else:
            paths.pop(version)
            paths.remove(ansys_path)

    version_float = float(version)/10.0
    return ansys_exe_path, version_float


def get_ansys_path(allow_input=True):
    """ Acquires ANSYS Path from a cached file or user input """
    exe_loc = None
    if os.path.isfile(CONFIG_FILE):
        with open(CONFIG_FILE) as f:
            exe_loc = f.read()
        # verify
        if not os.path.isfile(exe_loc) and allow_input:
            exe_loc = save_ansys_path()
    elif allow_input:  # create configuration file
        exe_loc = save_ansys_path()

    return exe_loc


def check_valid_ansys():
    """ Checks if a valid version of ANSYS is installed and preconfigured """
    ansys_bin = get_ansys_path(allow_input=False)
    if ansys_bin is not None:
        version = int(re.findall(r'\d\d\d', ansys_bin)[0])
        return not(version < 170 and os.name != 'posix')

    return False


def change_default_ansys_path(exe_loc):
    """Change your default ansys path.

    Parameters
    ----------
    exe_loc : str
        Ansys executable path.  Must be a full path.

    Examples
    --------
    Change default ansys location on Linux

    >>> import pyansys
    >>> pyansys.change_default_ansys_path('/ansys_inc/v201/ansys/bin/ansys201')
    >>> pyansys.mapdl.get_ansys_path()
    '/ansys_inc/v201/ansys/bin/ansys201'

    Change default ansys location on Windows
    >>> ans_pth = 'C:/Program Files/ANSYS Inc/v193/ansys/bin/win64/ANSYS193.exe'
    >>> pyansys.change_default_ansys_path(ans_pth)
    >>> pyansys.mapdl.check_valid_ansys()
    True

    """
    if os.path.isfile(exe_loc):
        with open(CONFIG_FILE, 'w') as f:
            f.write(exe_loc)
    else:
        raise FileNotFoundError('File %s is invalid or does not exist' % exe_loc)


def save_ansys_path(exe_loc=''):
    """ Find ANSYS path or query user """
    if exe_loc.strip():
        print('Cached ANSYS executable %s not found' % exe_loc)
    else:
        print('Cached ANSYS executable not found')
    exe_loc, ver = find_ansys()
    if os.path.isfile(exe_loc):
        print('Found ANSYS at %s' % exe_loc)
        resp = input('Use this location?  [Y/n]')
        if resp != 'n':
            change_default_ansys_path(exe_loc)
            return exe_loc

    if exe_loc is not None:
        if os.path.isfile(exe_loc):
            return exe_loc

    # otherwise, query user for the location
    with open(CONFIG_FILE, 'w') as f:
        try:
            exe_loc = raw_input('Enter location of ANSYS executable: ')
        except NameError:
            exe_loc = input('Enter location of ANSYS executable: ')
        if not os.path.isfile(exe_loc):
            raise FileNotFoundError('ANSYS executable not found at this location:\n%s' % exe_loc)

        f.write(exe_loc)

    return exe_loc


def check_lock_file(path, jobname, override):
    # Check for lock file
    lockfile = os.path.join(path, jobname + '.lock')
    if os.path.isfile(lockfile):
        if not override:
            raise LockFileException('\nLock file exists for jobname "%s"' % jobname +
                                    ' at\n"%s"\n\n' % lockfile +
                                    'Set ``override=True`` to or delete the lock file '
                                    'to start MAPDL')
        else:
            try:
                os.remove(lockfile)
            except PermissionError:
                raise LockFileException('Unable to remove lock file.  '
                                        'Another instance of MAPDL might be '
                                        'running at "%s"' % path)


def launch_mapdl(exec_file=None, run_location=None, mode=None, jobname='file',
                 nproc=2, override=False, loglevel='INFO',
                 additional_switches='', start_timeout=120,
                 log_apdl=False, **kwargs):
    """Launch a local instance of MAPDL in the background
    and allows commands to be passed to a persistent session.

    Parameters
    ----------
    exec_file : str, optional
        The location of the MAPDL executable.  Will use the cached
        location when left at the default ``None``.

    run_location : str, optional
        MAPDL working directory.  Defaults to a temporary working
        directory.

    mode : str, optional
        Mode to launch MAPDL.  Must be one of the following:

        - ``'corba'``
        - ``'console'``

    jobname : str, optional
        MAPDL jobname.  Defaults to ``'file'``.

    nproc : int, optional
        Number of processors.  Defaults to 2.

    override : bool, optional
        Attempts to delete the lock file at the run_location.
        Useful when a prior MAPDL session has exited prematurely and
        the lock file has not been deleted.

    wait : bool, optional
        When True, waits until MAPDL has been initialized before
        initializing the python ansys object.  Set this to False for
        debugging.

    loglevel : str, optional
        Sets which messages are printed to the console.  Default
        'INFO' prints out all MAPDL messages, 'WARNING` prints only
        messages containing MAPDL warnings, and 'ERROR' prints only
        error messages.

    additional_switches : str, optional
        Additional switches for MAPDL, for example aa_r, and academic
        research license, would be added with:

        - additional_switches="-aa_r"

        Avoid adding switches like -i -o or -b as these are already
        included to start up the MAPDL server.  See the notes
        section for additional details.

    start_timeout : float, optional
        Time to wait before raising an error that MAPDL is unable to
        start.

    log_broadcast : bool, optional
        Additional logging for ansys solution progress.  Default True
        and visible at log level 'INFO'.  Only applicable when ``mode=corba``

    log_apdl : str, optional
        Opens an APDL log file in the current MAPDL working directory
        to log all MAPDL commands to file.  Default False.  Set to 'a'
        to append to an existing log or 'w' to write a new log.

    Examples
    --------
    Launch MAPDL using the default configuration.

    >>> import pyansys
    >>> mapdl = pyansys.Mapdl()

    Run MAPDL with shared memory parallel and specify the location of
    the ansys binary.

    >>> exec_file = 'C:/Program Files/ANSYS Inc/v201/ansys/bin/win64/ANSYS201.exe'
    >>> mapdl = pyansys.Mapdl(exec_file, additional_switches='-smp')

    Run MAPDL using the console mode (only on linux).

    >>> mapdl = pyansys.Mapdl('/ansys_inc/v194/ansys/bin/ansys194',
                              mode='console')

    Notes
    -----
    MAPDL has the following command line options as of 2020R1

    -aas : Enables server mode
     When enabling server mode, a custom name for the keyfile can be
     specified using the ``-iorFile`` option.  This is the CORBA that
     pyansys uses for ``mode='corba'``.

    -acc <device> : Enables the use of GPU hardware.  Enables the use of
     GPU hardware to accelerate the analysis. See GPU Accelerator
     Capability in the Parallel Processing Guide for more information.

    -amfg : Enables the additive manufacturing capability.
     Requires an additive manufacturing license. For general
     information about this feature, see AM Process Simulation in
     ANSYS Workbench.

    -ansexe <executable> :  activates a custom mechanical APDL executable.
     In the ANSYS Workbench environment, activates a custom
     Mechanical APDL executable.

    -b <list or nolist> : Activates batch mode
     The options ``-b`` list or ``-b`` by itself cause the input
     listing to be included in the output. The ``-b`` nolist option
     causes the input listing not to be included. For more information
     about running Mechanical APDL in batch mode, see Batch Mode.

    -custom <executable> : Calls a custom Mechanical APDL executable
     See Running Your Custom Executable in the Programmer's Reference
     for more information.

    -d <device> : Specifies the type of graphics device
     This option applies only to interactive mode. For Linux systems,
     graphics device choices are X11, X11C, or 3D. For Windows
     systems, graphics device options are WIN32 or WIN32C, or 3D.

    -db value : Initial memory allocation
     Defines the portion of workspace (memory) to be used as the
     initial allocation for the database. The default is 1024
     MB. Specify a negative number to force a fixed size throughout
     the run; useful on small memory systems.

    -dir <path> : Defines the initial working directory
     Using the ``-dir`` option overrides the
     ``ANSYS201_WORKING_DIRECTORY`` environment variable.

    -dis : Enables Distributed ANSYS
     See the Parallel Processing Guide for more information.

    -dvt : Enables ANSYS DesignXplorer advanced task (add-on).
     Requires DesignXplorer.

    -g : Launches the Mechanical APDL program with the GUI
     Graphical User Interface (GUI) on. If you select this option, an
     X11 graphics device is assumed for Linux unless the ``-d`` option
     specifies a different device. This option is not used on Windows
     systems. To activate the GUI after Mechanical APDL has started,
     enter two commands in the input window: /SHOW to define the
     graphics device, and /MENU,ON to activate the GUI. The ``-g`` option
     is valid only for interactive mode.  Note: If you start
     Mechanical APDL via the ``-g`` option, the program ignores any /SHOW
     command in the start.ans file and displays a splash screen
     briefly before opening the GUI windows.

    -i <inputname> : Specifies the name of the file to read
     Inputs an input file into Mechanical APDL for batch
     processing.

    -iorFile <keyfile_name> : Specifies the name of the server keyfile
     Name of the server keyfile when enabling server mode. If this
     option is not supplied, the default name of the keyfile is
     ``aas_MapdlID.txt``. For more information, see Mechanical APDL as
     a Server Keyfile in the Mechanical APDL as a Server User's Guide.

    -j <Jobname> : Specifies the initial jobname
     A name assigned to all files generated by the program for a
     specific model. If you omit the ``-j`` option, the jobname is assumed
     to be file.

    -l <language> : Specifies a language file to use other than English
     This option is valid only if you have a translated message file
     in an appropriately named subdirectory in
     ``/ansys_inc/v201/ansys/docu`` or
     ``Program Files\\ANSYS\\Inc\\V201\\ANSYS\\docu``

    -m <workspace> : Specifies the total size of the workspace
     Workspace (memory) in megabytes used for the initial
     allocation. If you omit the ``-m`` option, the default is 2 GB
     (2048 MB). Specify a negative number to force a fixed size
     throughout the run.

    -machines <IP> : Specifies the distributed machines
     Machines on which to run a Distributed ANSYS analysis. See
     Starting Distributed ANSYS in the Parallel Processing Guide for
     more information.

    -mpi <value> : Specifies the type of MPI to use.
     See the Parallel Processing Guide for more information.

    -mpifile <appfile> : Specifies an existing MPI file
     Specifies an existing MPI file (appfile) to be used in a
     Distributed ANSYS run. See Using MPI Files in the Parallel
     Processing Guide for more information.

    -na <value>: Specifies the number of GPU accelerator devices
     Nubmer of GPU devices per machine or compute node when running
     with the GPU accelerator feature. See GPU Accelerator Capability
     in the Parallel Processing Guide for more information.

    -name <value> : Defines Mechanical APDL parameters
     Set mechanical APDL parameters at program start-up. The parameter
     name must be at least two characters long. For details about
     parameters, see the ANSYS Parametric Design Language Guide.

    -np <value> : Specifies the number of processors
     Number of processors to use when running Distributed ANSYS or
     Shared-memory ANSYS. See the Parallel Processing Guide for more
     information.

    -o <outputname> : Output file
     Specifies the name of the file to store the output from a batch
     execution of Mechanical APDL

    -p <productname> : ANSYS session product
     Defines the ANSYS session product that will run during the
     session. For more detailed information about the ``-p`` option,
     see Selecting an ANSYS Product via the Command Line.

    -ppf <license feature name> : HPC license
     Specifies which HPC license to use during a parallel processing
     run. See HPC Licensing in the Parallel Processing Guide for more
     information.

    -rcopy <path> : Path to remote copy of files
     On a Linux cluster, specifies the full path to the program used
     to perform remote copy of files. The default value is
     ``/usr/bin/scp``.

    -s <read or noread> : Read startup file
     Specifies whether the program reads the ``start.ans`` file at
     start-up. If you omit the ``-s`` option, Mechanical APDL reads the
     start.ans file in interactive mode and not in batch mode.

    -schost <host>: Coupling service host
     Specifies the host machine on which the coupling service is
     running (to which the co-simulation participant/solver must
     connect) in a System Coupling analysis.

    -scid <value> : System Coupling analysis licensing ID
     Specifies the licensing ID of the System Coupling analysis.

    -sclic <port@host> : System Coupling analysis host
     Specifies the licensing port and host to use for the System
     Coupling analysis.

    -scname <solver> : Name of the co-simulation participant
     Specifies the unique name used by the co-simulation participant
     to identify itself to the coupling service in a System Coupling
     analysis. For Linux systems, you need to quote the name to have
     the name recognized if it contains a space: (e.g.
     ``-scname "Solution 1"``)

    -scport <port> : Coupling service port
     Specifies the port on the host machine upon which the coupling
     service is listening for connections from co-simulation
     participants in a System Coupling analysis.

    -smp : Enables shared-memory parallelism.
     See the Parallel Processing Guide for more information.

    -usersh : Enable MPI remote shell.
     Directs the MPI software (used by Distributed ANSYS) to use the
     remote shell (rsh) protocol instead of the default secure shell
     (ssh) protocol. See Configuring Distributed ANSYS in the Parallel
     Processing Guide for more information.

    -v : Return version info
     Returns the Mechanical APDL release number, update number,
     copyright date, customer number, and license manager version
     number.
    """
    # depreciated options
    if 'prefer_pexpect' in kwargs:
        raise NotImplementedError('"prefer_pexpect" is depreciated.  '
                                  'Please use: ``mode="console"``')

    if exec_file is None:
        # Load cached path
        exec_file = get_ansys_path()
        if exec_file is None:
            raise FileNotFoundError('Invalid or path or cannot load cached '
                                    'ansys path.  Enter one manually using '
                                    'launch_mapdl(exec_file=)')
    else:  # verify ansys exists at this location
        if not os.path.isfile(exec_file):
            raise FileNotFoundError('Invalid ANSYS executable at "%s"\n'
                                    % exec_file + 'Enter one manually using '
                                    'launch_mapdl(exec_file=)')

    if run_location is None:
        temp_dir = tempfile.gettempdir()
        run_location = os.path.join(temp_dir, 'ansys_%s' % random_string(10))
        if not os.path.isdir(run_location):
            try:
                os.mkdir(run_location)
            except:
                raise RuntimeError('Unable to create temporary working '
                                   'directory %s\n' % run_location +
                                   'Please specify run_location=')
    else:
        if not os.path.isdir(run_location):
            raise FileNotFoundError('"%s" is not a valid directory' % run_location)

    check_lock_file(run_location, jobname, override)

    # NOTE: version may or may not be within the full exec_path
    version = int(re.findall(r'\d\d\d', exec_file)[0])
    mode = check_mode(mode, version)

    # known issue with distributed memory parallel
    if 'ubuntu' in platform.platform().lower():
        if 'smp' not in additional_switches:
            # Ubuntu ANSYS fails to launch without I_MPI_SHM_LMT
            os.environ['I_MPI_SHM_LMT'] = 'shm'

    start_parm = {'exec_file': exec_file,
                  'run_location': run_location,
                  'jobname': jobname,
                  'nproc': nproc,
                  'additional_switches': additional_switches,
                  'start_timeout': start_timeout}

    if mode == 'console':
        from pyansys.mapdl_console import MapdlConsole
        return MapdlConsole(loglevel=loglevel, log_apdl=log_apdl, **start_parm)
    elif mode == 'corba':
        from pyansys.mapdl_corba import MapdlCorba
        return MapdlCorba(loglevel=loglevel, log_apdl=log_apdl,
                          log_broadcast=kwargs.get('log_broadcast',
                                                   False), **start_parm)
    else:
        raise ValueError('Invalid mode %s' % mode)


def check_mode(mode, version):
    """Check if the MAPDL server mode matches the allowable version

    If ``None``, the newest mode will be selected.
    """

    is_win = os.name == 'nt'

    if isinstance(mode, str):
        mode = mode.lower()
        if mode not in ALLOWABLE_MODES:
            raise ValueError('Invalid MAPDL server mode.  '
                             'Use one of the following:\n%s' % ALLOWABLE_MODES)

        if mode == 'corba':
            if version < 170:
                raise VersionError('CORBA AAS mode requires MAPDL v17.0 or newer.')

        elif mode == 'console' and is_win:
            raise ValueError('Console mode requires Linux')

    else:  # auto-select based on best version
        if version >= 170:
            mode = 'corba'
        else:
            if os.name == 'nt':
                raise VersionError('Running MAPDL as a service requires '
                                   'v17.0 or greater on Windows.')
            mode = 'console'

    if version < 130:
        warnings.warn('MAPDL as a service has not been tested on MAPDL < v13')

    return mode



import re


def parse_e(msg):
    """Parse create element message and return element number"""
    res = re.search(r"(ELEMENT\s*)([0-9]+)", msg)
    if res is not None:
        result = int(res.group(2))
    else:
        result = None
    return result


def parse_et(msg):
    """
    Parse local element type number definition message
    and return element type number
    """
    res = re.search(r"(ELEMENT TYPE\s*)([0-9]+)", msg)
    if res is not None:
        result = int(res.group(2))
    else:
        result = None
    return result


element_commands = {'E': parse_e,
                    'ET': parse_et,
}



"""Module for common class between Archive, and Result mesh"""
import pyvista as pv
import vtk
import numpy as np

from pyansys import _relaxmidside, _reader
from pyansys.misc import unique_rows
from pyansys.elements import ETYPE_MAP


VTK9 = vtk.vtkVersion().GetVTKMajorVersion() >= 9
INVALID_ALLOWABLE_TYPES = TypeError('`allowable_types` must be an array '
                                    'of ANSYS element types from 1 and 300')

# map MESH200 elements to a pyansys/VTK element type (see elements.py)
MESH200_MAP = {0: 2,  # line
               1: 2,  # line
               2: 2,  # line
               3: 2,  # line
               4: 3,  # triangle
               5: 3,  # triangle
               6: 3,  # quadrilateral
               7: 3,  # quadrilateral
               8: 5,  # tetrahedron with 4 nodes
               9: 5,  # tetrahedron with 10 nodes
               10: 4,  # hex with 8 nodes
               11: 4}  # hex with 8 nodes


class Mesh():

    def __init__(self, nnum=None, nodes=None, elem=None,
                 elem_off=None, ekey=None, node_comps={},
                 elem_comps={}, rdat=[], rnum=[], keyopt={}):
        self._etype = None  # internal element type reference
        self._grid = None  # VTK grid
        self._surf_cache = None # cached external surface
        self._enum = None  # cached element numbering
        self._etype_cache = None  # cached ansys element type numbering
        self._rcon = None  # cached ansys element real constant
        self._mtype = None  # cached ansys material type
        self._node_angles = None  # cached node angles
        self._node_coord = None  # cached node coordinates
        self._cached_elements = None  # cached list of elements
        self._secnum = None  # cached section number
        self._esys = None  # cached element coordinate system

        # Always set on init
        self._nnum = nnum
        self._nodes = nodes
        self._elem = elem
        self._elem_off = elem_off
        self._ekey = ekey

        # optional
        self._node_comps = node_comps
        self._elem_comps = elem_comps
        self._rdat = rdat
        self._rnum = rnum
        self._keyopt = keyopt

    @property
    def _surf(self):
        """External surface"""
        if self._surf_cache is None:
            self._surf_cache = self._grid.extract_surface()
        return self._surf_cache

    @property
    def _has_nodes(self):
        """Returns True when has nodes"""
        # if isinstance(self._nodes, np.ndarray):
            # return bool(self._nodes.size)
        return len(self.nodes)

    @property
    def _has_elements(self):
        """Returns True when geometry has elements"""
        if self._elem is None:
            return False

        if isinstance(self._elem, np.ndarray):
            return self._elem.size

        return len(self._elem)

    def _parse_vtk(self, allowable_types=None, force_linear=False,
                   null_unallowed=False, fix_midside=True, additional_checking=False):
        """Convert raw ANSYS nodes and elements to a VTK UnstructuredGrid

        Parameters
        ----------
        fix_midside : bool, optional
            Adds additional midside nodes when ``True``.  When
            ``False``, missing ANSYS cells will simply point to the
            first node.

        """
        if not self._has_nodes or not self._has_elements:
            # warnings.warn('Missing nodes or elements.  Unable to parse to vtk')
            return

        etype_map = ETYPE_MAP
        if allowable_types is not None:
            try:
                allowable_types = np.asarray(allowable_types)
            except:
                raise INVALID_ALLOWABLE_TYPES

            if not issubclass(allowable_types.dtype.type, np.integer):
                raise TypeError('Element types must be an integer array-like')

            if allowable_types.min() < 1 or allowable_types.max() > 300:
                raise INVALID_ALLOWABLE_TYPES

            etype_map = np.zeros_like(ETYPE_MAP)
            etype_map[allowable_types] = ETYPE_MAP[allowable_types]

        # ANSYS element type to VTK map
        type_ref = np.empty(2 << 15, np.int32)  # 65536
        type_ref[self._ekey[:, 0]] = etype_map[self._ekey[:, 1]]        

        # special treatment for MESH200
        if allowable_types is None or 200 in allowable_types:
            for etype_ind, etype in self._ekey:
                if etype == 200 and etype_ind in self.key_option:
                    # keyoption 1 contains various cell types
                    # map them to the corresponding type (see elements.py)
                    mapped = MESH200_MAP[self.key_option[etype_ind][0][1]]
                    type_ref[etype_ind] = mapped

        offset, celltypes, cells = _reader.ans_vtk_convert(self._elem,
                                                           self._elem_off,
                                                           type_ref,
                                                           self.nnum,
                                                           True)  # for reset_midside

        nodes, angles, nnum = self.nodes, self.node_angles, self.nnum

        # fix missing midside
        if np.any(cells == -1):
            if fix_midside:
                nodes, angles, nnum = fix_missing_midside(cells, nodes, celltypes,
                                                          offset, angles, nnum)
            else:
                cells[cells == -1] = 0

        if additional_checking:
            cells[cells < 0] = 0
            # cells[cells >= nodes.shape[0]] = 0  # fails when n_nodes < 20

        if VTK9:
            grid = pv.UnstructuredGrid(cells, celltypes, nodes, deep=False)
        else:
            grid = pv.UnstructuredGrid(offset, cells, celltypes, nodes,
                                       deep=False)

        # Store original ANSYS element and node information
        grid.point_arrays['ansys_node_num'] = nnum
        grid.cell_arrays['ansys_elem_num'] = self.enum
        grid.cell_arrays['ansys_real_constant'] = self.elem_real_constant
        grid.cell_arrays['ansys_material_type'] = self.material_type
        grid.cell_arrays['ansys_etype'] = self._ans_etype
        grid.cell_arrays['ansys_elem_type_num'] = self.etype

        # add components
        # Add element components to unstructured grid
        for key, item in self.element_components.items():
            mask = np.in1d(self.enum, item, assume_unique=True)
            grid.cell_arrays[key] = mask

        # Add node components to unstructured grid
        for key, item in self.node_components.items():
            mask = np.in1d(nnum, item, assume_unique=True)
            grid.point_arrays[key] = mask

        # store node angles
        if angles is not None:
            if angles.shape[1] == 3:
                grid.point_arrays['angles'] = angles

        if not null_unallowed:
            grid = grid.extract_cells(grid.celltypes != 0)

        if force_linear:
            grid = grid.linear_copy()

        # map over element types
        # Add tracker for original node numbering
        ind = np.arange(grid.number_of_points)
        grid.point_arrays['origid'] = ind
        grid.point_arrays['VTKorigID'] = ind
        return grid

    @property
    def key_option(self):
        """Additional key options for element types

        Examples
        --------
        >>> import pyansys
        >>> archive = pyansys.Archive(pyansys.examples.hexarchivefile)
        >>> archive.key_option
        {1: [[1, 11]]}
        """
        return self._keyopt

    @property
    def material_type(self):
        """Material type index of each element in the archive.

        Examples
        --------
        >>> import pyansys
        >>> archive = pyansys.Archive(pyansys.examples.hexarchivefile)
        >>> archive.material_type
        array([1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
               1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
               1, 1, 1, 1], dtype=int32)
        """
        # FIELD 0 : material reference number
        if self._mtype is None:
            self._mtype = self._elem[self._elem_off[:-1]]
        return self._mtype

    @property
    def element_components(self):
        """Element components for the archive.

        Output is a dictionary of element components.  Each entry is an
        array of MAPDL element numbers corresponding to the element
        component.  The keys are element component names.

        Examples
        --------
        >>> import pyansys
        >>> archive = pyansys.Archive(pyansys.examples.hexarchivefile)
        >>> archive.element_components
        {'ECOMP1 ': array([17, 18, 21, 22, 23, 24, 25, 26, 27, 28, 29,
                           30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40],
                           dtype=int32),
        'ECOMP2 ': array([ 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13,
                          14, 15, 16, 17, 18, 19, 20, 23, 24], dtype=int32)}
        """
        return self._elem_comps

    @property
    def node_components(self):
        """Node components for the archive.

        Output is a dictionary of node components.  Each entry is an
        array of MAPDL node numbers corresponding to the node
        component.  The keys are node component names.

        Examples
        --------
        >>> import pyansys
        >>> archive = pyansys.Archive(pyansys.examples.hexarchivefile)
        >>> archive.node_components
        {'NCOMP2  ': array([  1,   2,   3,   4,   5,   6,   7,   8,
                             14, 15, 16, 17, 18, 19, 20, 21, 43, 44,
                             62, 63, 64, 81, 82, 90, 91, 92, 93, 94,
                             118, 119, 120, 121, 122, 123, 124, 125,
                             126, 137, 147, 148, 149, 150, 151, 152,
                             153, 165, 166, 167, 193, 194, 195, 202,
                             203, 204, 205, 206, 207, 221, 240, 258,
                             267, 268, 276, 277, 278, 285, 286, 287,
                             304, 305, 306, 313, 314, 315, 316
                             ], dtype=int32),
        ...,
        }
        """
        return self._node_comps

    @property
    def elem_real_constant(self):
        """Real constant reference for each element.

        Use the data within ``rlblock`` and ``rlblock_num`` to get the
        real constant datat for each element.

        Examples
        --------
        >>> import pyansys
        >>> archive = pyansys.Archive(pyansys.examples.hexarchivefile)
        >>> archive.elem_real_constant
        array([ 1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,
                1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,
                1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,
                ...,
                1,  1,  1,  1,  1,  1,  1,  1,  1,  1, 61, 61, 61, 61,
               61, 61, 61, 61, 61, 61, 61, 61, 61, 61, 61, 61, 61, 61,
               61], dtype=int32)

        """
        # FIELD 2 : real constant reference number
        if self._rcon is None:
            self._rcon = self._elem[self._elem_off[:-1] + 2]
        return self._rcon

    @property
    def etype(self):
        """Element type of each element.

        This is the ansys element type for each element.

        Examples
        --------
        >>> import pyansys
        >>> archive = pyansys.Archive(pyansys.examples.hexarchivefile)
        >>> archive.etype
        array([ 45,  45,  45,  45,  45,  45,  45,  45,  45,  45,  45,
                45,  45,  45,  45,  45,  45,  45,  45,  92,  92,  92,
                92,  92,  92,  92,  92,  92,  92,  92,  92,  92,  92,
                ...,
                92,  92,  92,  92,  92, 154, 154, 154, 154, 154, 154,
               154, 154, 154, 154, 154, 154, 154, 154, 154, 154, 154,
               154], dtype=int32)

        Notes
        -----
        Element types are listed below.  Please see the APDL Element
        Reference for more details:

        https://www.mm.bme.hu/~gyebro/files/vem/ansys_14_element_reference.pdf
        """
        if self._etype is None:
            arr = np.empty(self._ekey[:, 0].max() + 1, np.int32)
            arr[self._ekey[:, 0]] = self._ekey[:, 1]
            self._etype = arr[self._ans_etype]
        return self._etype

    @property
    def _ans_etype(self):
        """FIELD 1 : element type number"""
        if self._etype_cache is None:
            self._etype_cache = self._elem[self._elem_off[:-1] + 1]
        return self._etype_cache

    @property
    def section(self):
        """Section number

        Examples
        --------
        >>> import pyansys
        >>> archive = pyansys.Archive(pyansys.examples.hexarchivefile)
        >>> archive.section
        array([1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
               1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
               1, 1], dtype=int32)
        """
        if self._secnum is None:
            self._secnum = self._elem[self._elem_off[:-1] + 3]  # FIELD 3
        return self._secnum

    def element_coord_system(self):
        """Element coordinate system number

        Examples
        --------
        >>> import pyansys
        >>> archive = pyansys.Archive(pyansys.examples.hexarchivefile)
        >>> archive.element_coord_system
        array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
               0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
               0, 0], dtype=int32)
        """
        if self._esys is None:
            self._esys = self._elem[self._elem_off[:-1] + 4]  # FIELD 4
        return self._esys

    @property
    def elem(self):
        """List of elements containing raw ansys information.

        Each element contains 10 items plus the nodes belonging to the
        element.  The first 10 items are:

        - FIELD 0 : material reference number
        - FIELD 1 : element type number
        - FIELD 2 : real constant reference number
        - FIELD 3 : section number
        - FIELD 4 : element coordinate system
        - FIELD 5 : death flag (0 - alive, 1 - dead)
        - FIELD 6 : solid model reference
        - FIELD 7 : coded shape key
        - FIELD 8 : element number
        - FIELD 9 : base element number (applicable to reinforcing elements only)
        - FIELDS 10 - 30 : The nodes belonging to the element in ANSYS numbering.

        Examples
        --------
        >>> import pyansys
        >>> archive = pyansys.Archive(pyansys.examples.hexarchivefile)
        >>> archive.elem
        [array([  1,   4,  19,  15,  63,  91, 286, 240,   3,  18,  17,
                 16,  81, 276, 267, 258,  62,  90, 285, 239],
         array([  4,   2,   8,  19,  91,  44, 147, 286,   5,   7,  21,
                 18, 109, 137, 313, 276,  90,  43, 146, 285],
         array([ 15,  19,  12,  10, 240, 286, 203, 175,  17,  20,  13,
                 14, 267, 304, 221, 230, 239, 285, 202, 174],
        ...
        """
        if self._cached_elements is None:
            self._cached_elements = np.split(self._elem, self._elem_off[1:-1])
        return self._cached_elements

    @property
    def enum(self):
        """ANSYS element numbers.

        Examples
        --------
        >>> import pyansys
        >>> archive = pyansys.Archive(pyansys.examples.hexarchivefile)
        >>> archive.enum
        array([    1,     2,     3, ...,  9998,  9999, 10000])
        """
        if self._enum is None:
            self._enum = self._elem[self._elem_off[:-1] + 8]
        return self._enum

    @property
    def nnum(self):
        """Array of node numbers.

        Examples
        --------
        >>> import pyansys
        >>> archive = pyansys.Archive(pyansys.examples.hexarchivefile)
        >>> archive.nnum
        array([    1,     2,     3, ..., 19998, 19999, 20000])
        """
        return self._nnum

    @property
    def ekey(self):
        """Element type key

        Array containing element type numbers in the first column and
        the element types (like SURF154) in the second column.

        Examples
        --------
        >>> import pyansys
        >>> archive = pyansys.Archive(pyansys.examples.hexarchivefile)
        >>> archive.ekey
        array([[  1,  45],
               [  2,  95],
               [  3,  92],
               [ 60, 154]], dtype=int32)
        """
        return self._ekey

    @property
    def rlblock(self):
        """Real constant data from the RLBLOCK.

        Examples
        --------
        >>> import pyansys
        >>> archive = pyansys.Archive(pyansys.examples.hexarchivefile)
        >>> archive.rlblock
        [[0.   , 0.   , 0.   , 0.   , 0.   , 0.   , 0.02 ],
         [0.   , 0.   , 0.   , 0.   , 0.   , 0.   , 0.01 ],
         [0.   , 0.   , 0.   , 0.   , 0.   , 0.   , 0.005],
         [0.   , 0.   , 0.   , 0.   , 0.   , 0.   , 0.005]]
        """
        return self._rdat

    @property
    def rlblock_num(self):
        """Indices from the real constant data

        Examples
        --------
        >>> import pyansys
        >>> archive = pyansys.Archive(pyansys.examples.hexarchivefile)
        >>> archive.rnum
        array([60, 61, 62, 63])
        """
        return self._rnum

    @property
    def nodes(self):
        """Array of nodes.

        Examples
        --------
        >>> import pyansys
        >>> archive = pyansys.Archive(pyansys.examples.hexarchivefile)
        >>> archive.nodes
        [[0.   0.   0.  ]
         [1.   0.   0.  ]
         [0.25 0.   0.  ]
         ...,
         [0.75 0.5  3.5 ]
         [0.75 0.5  4.  ]
         [0.75 0.5  4.5 ]]
        """
        if self._node_coord is None:
            self._node_coord = np.ascontiguousarray(self._nodes[:, :3])
        return self._node_coord

    @property
    def node_angles(self):
        """Node angles from the archive file.

        Examples
        --------
        >>> import pyansys
        >>> archive = pyansys.Archive(pyansys.examples.hexarchivefile)
        >>> archive.nodes
        [[0.   0.   0.  ]
         [0.   0.   0.  ]
         [0.   0.   0.  ]
         ...,
         [0.   0.   0.  ]
         [0.   0.   0.  ]
         [0.   0.   0.  ]]
        """
        if self._node_angles is None:
            self._node_angles = np.ascontiguousarray(self._nodes[:, 3:])
        return self._node_angles

    def __repr__(self):
        txt = 'ANSYS Mesh\n'
        txt += '  Number of Nodes:              %d\n' % len(self.nnum)
        txt += '  Number of Elements:           %d\n' % len(self.enum)
        txt += '  Number of Element Types:      %d\n' % len(self.ekey)
        txt += '  Number of Node Components:    %d\n' % len(self.node_components)
        txt += '  Number of Element Components: %d\n' % len(self.element_components)
        return txt

    def save(self, filename, binary=True, force_linear=False, allowable_types=[],
             null_unallowed=False):
        """Save the geometry as a vtk file

        Parameters
        ----------
        filename : str
            Filename of output file. Writer type is inferred from
            the extension of the filename.

        binary : bool, optional
            If ``True``, write as binary, else ASCII.

        force_linear : bool, optional
            This parser creates quadratic elements if available.  Set
            this to True to always create linear elements.  Defaults
            to False.

        allowable_types : list, optional
            Allowable element types.  Defaults to all valid element
            types in ``pyansys.elements.valid_types``

            See ``help(pyansys.elements)`` for available element types.

        null_unallowed : bool, optional
            Elements types not matching element types will be stored
            as empty (null) elements.  Useful for debug or tracking
            element numbers.  Default False.

        Examples
        --------
        >>> geom.save('mesh.vtk')

        Notes
        -----
        Binary files write much faster than ASCII and have a smaller
        file size.
        """
        grid = self._parse_vtk(allowable_types=allowable_types,
                               force_linear=force_linear,
                               null_unallowed=null_unallowed)
        return grid.save(filename, binary=binary)

    @property
    def n_node(self):
        """Number of nodes"""
        if not self._has_nodes:
            return 0
        return self.nodes.shape[0]

    @property
    def n_elem(self):
        """Number of nodes"""
        if not self._has_elements:
            return 0
            
        return len(self.enum)


def fix_missing_midside(cells, nodes, celltypes, offset, angles, nnum):
    """Adds missing midside nodes to cells.

    ANSYS sometimes does not add midside nodes, and this is denoted in
    the element array with a ``0``.  When translated to VTK, this is
    saved as a ``-1``.  If this is not corrected, VTK will segfault.

    This function creates missing midside nodes for the quadratic
    elements.
    """
    # Check for missing midside nodes
    mask = cells == -1
    nnodes = nodes.shape[0]

    nextra = mask.sum()
    cells[mask] = np.arange(nnodes, nnodes + nextra)

    nodes_new = np.empty((nnodes + nextra, 3))
    nodes_new[:nnodes] = nodes

    # Set new midside nodes directly between their edge nodes
    temp_nodes = nodes_new.copy()
    _relaxmidside.reset_midside(cells, celltypes, offset, temp_nodes)

    # merge midside nodes
    unique_nodes, idx_a, idx_b = unique_rows(temp_nodes[nnodes:])

    # rewrite node numbers
    cells[mask] = idx_b + nnodes
    nextra = idx_a.shape[0]  # extra unique nodes
    nodes_new = nodes_new[:nnodes + nextra]
    nodes_new[nnodes:] = unique_nodes

    if angles is not None:
        new_angles = np.empty((nnodes + nextra, 3))
        new_angles[:nnodes] = angles
        new_angles[nnodes:] = 0
    else:
        new_angles = None

    # Add extra node numbers
    nnum_new = np.empty(nnodes + nextra)
    nnum_new[:nnodes] = nnum
    nnum_new[nnodes:] = -1
    return nodes_new, new_angles, nnum_new



LOCKFILE_MSG = """
Another ANSYS job with the same job name is already running in this
directory, or the lock file has not been deleted from an abnormally
terminated ANSYS run.

Disable this check by passing ``override=True``

"""

class VersionError(ValueError):
    """Raised when MAPDL is the wrong version"""

    def __init__(self, msg='Invalid MAPDL version'):
        ValueError.__init__(self, msg)


class NoDistributedFiles(FileNotFoundError):
    """Unable to find any distributed result files """

    def __init__(self, msg='Unable to find any distributed result files'):
        FileNotFoundError.__init__(self, msg)


class MapdlRuntimeError(RuntimeError):
    """Raised when MAPDL passes an error"""
    pass


class MapdlInvalidRoutineError(RuntimeError):
    """Raised when MAPDL is in the wrong routine"""
    def __init__(self, msg=''):
        RuntimeError.__init__(self, msg)


class MapdlExitedError(RuntimeError):
    """Raised when MAPDL has exited"""

    def __init__(self, msg='MAPDL has exited'):
        RuntimeError.__init__(self, msg)


class LockFileException(RuntimeError):
    """Error message when the lockfile has not been removed"""
    def __init__(self, msg=LOCKFILE_MSG):
        RuntimeError.__init__(self, msg)



"""Supports reading cyclic structural result files from ANSYS"""
from functools import wraps

from vtk import vtkMatrix4x4, vtkTransform, vtkAppendFilter
import numpy as np
from pyvista.core.common import axis_rotation
import pyvista as pv

from pyansys.common import (STRESS_TYPES, STRAIN_TYPES,
                            PRINCIPAL_STRESS_TYPES,
                            THERMAL_STRAIN_TYPES)
from pyansys.rst import Result, check_comp
from pyansys import _binary_reader

np.seterr(divide='ignore', invalid='ignore')


class CyclicResult(Result):
    """Adds cyclic functionality to the result class"""

    def __init__(self, filename):
        """Initializes cyclic result"""
        super().__init__(filename)

        # sanity check
        if not self._is_cyclic:
            raise TypeError('Result is not a cyclic model')

        self._rotor_cache = None
        self._has_duplicate_sector = None
        self._is_repeated_mode = np.empty(0)
        self._repeated_index = np.empty(0)

        self._add_cyclic_properties()

    def plot_sectors(self, **kwargs):
        """Plot the full rotor and individually color the sectors.

        Parameters
        ----------
        kwargs : keyword arguments
            Additional keyword arguments.  See ``help(pyvista.plot)``

        Examples
        --------
        >>> from pyansys import examples
        >>> rst = examples.download_academic_rotor()
        >>> rst.plot_sectors()

        Save a screenshot of the sectors

        >>> rst.plot_sectors(screenshot='sectors.png')

        """
        scalars = np.empty((self.n_sector, self._mas_grid.n_points))
        scalars[:] = np.arange(self.n_sector).reshape(-1, 1)
        kwargs.setdefault('show_edges', True)
        kwargs.setdefault('n_colors', self.n_sector)
        return self._plot_cyclic_point_scalars(scalars, None, add_text=False, **kwargs)

    def plot(self, **kwargs):
        """Plot the full rotor geometry.

        Parameters
        ----------
        kwargs : keyword arguments
            Additional keyword arguments.  See ``help(pyvista.plot)``

        Returns
        -------
        cpos : list
            List of camera position, focal point, and view up.

        Examples
        --------
        >>> from pyansys import examples
        >>> rst = examples.download_academic_rotor()
        >>> rst.plot()

        Save a screenshot of the rotor

        >>> rst.plot(screenshot='rotor.png')

        """
        kwargs.setdefault('color', 'w')
        kwargs.setdefault('show_edges', True)
        return self._plot_cyclic_point_scalars(None, None, add_text=False, **kwargs)

    def _add_cyclic_properties(self):
        """Add cyclic properties"""

        # idenfity the sector based on number of elements in master sector
        cs_els = self._resultheader['csEls']
        mask = self.quadgrid.cell_arrays['ansys_elem_num'] <= cs_els

        self.master_cell_mask = mask
        self._mas_grid = self.grid.extract_cells(mask)

        # NOTE: number of nodes in sector may not match number of
        # nodes in geometry
        node_mask = self._neqv[self._sidx] <= self._resultheader['csNds']
        self._mas_ind = np.nonzero(node_mask)[0]
        self._dup_ind = np.nonzero(~node_mask)[0]
        self._has_duplicate_sector = np.any(self._dup_ind)

        # determine repeated modes
        mask_a = np.isclose(self.time_values, np.roll(self.time_values, 1))
        mask_b = np.isclose(self.time_values, np.roll(self.time_values, -1))
        self._is_repeated_mode = np.logical_or(mask_a, mask_b)

        # edge case for single pair of repeated modes
        if self._is_repeated_mode.size == 2 and self._is_repeated_mode.all():
            self._repeated_index = np.array([1, 0])
            return
        elif self._is_repeated_mode.size == 1:  # edge case single result
            self._is_repeated_mode = np.array([False])
            return

        self._repeated_index = np.empty(self._is_repeated_mode.size, np.int)
        self._repeated_index[:] = -1
        if np.any(self._is_repeated_mode):
            self._repeated_index[mask_a] = np.nonzero(mask_b)[0]
            self._repeated_index[mask_b] = np.nonzero(mask_a)[0]

    def nodal_solution(self, rnum, phase=0, full_rotor=False, as_complex=False):
        """Returns the DOF solution for each node in the global
        cartesian coordinate system.

        Parameters
        ----------
        rnum : int or list
            Cumulative result number with zero based indexing, or a
            list containing (step, substep) of the requested result.

        phase : float, optional
            Phase to rotate sector result in radians.

        full_rotor : bool, optional
            Expands the single sector solution for the full rotor.
            Sectors are rotated counter-clockwise about the axis of
            rotation.  Default False.

        as_complex : bool, optional
            Returns result as a complex number, otherwise as the real
            part rotated by phase.  Default False.

        Returns
        -------
        nnum : numpy.ndarray
            Node numbers of master sector.

        result : numpy.ndarray
            Result is (nnod x numdof), nnod is the number of nodes in
            a sector and numdof is the number of degrees of freedom.
            When full_rotor is True the array will be (nSector x nnod
            x numdof).

        Examples
        --------
        Visualize the 1st nodal diameter mode.

        >>> import pyansys
        >>> result = pyansys.download_academic_rotor()
        >>> result.nodal_solution((2, 1))

        Same result but uses Python (zero based) cumulative indexing

        >>> result.nodal_solution(2)

        Notes
        -----
        Somewhere between v15.0 and v18.2 ANSYS stopped writing the
        duplicate sector to the result file and instead records results in
        pairs (i.e. harmonic index 1, -1).

        """

        func = super().nodal_solution
        return self._get_full_result(rnum, func, phase, full_rotor, as_complex,
                                     tensor=False)

    @wraps(nodal_solution)
    def nodal_displacement(self, *args, **kwargs):
        """wraps nodal_solution"""
        return self.nodal_solution(*args, **kwargs)

    def _expand_cyclic_static(self, result, tensor=False, stress=True):
        """Expand cyclic static result for a full rotor"""
        cs_cord = self._resultheader['csCord']
        if cs_cord > 1:
            matrix = self.cs_4x4(cs_cord, as_vtk_matrix=True)
            i_matrix = self.cs_4x4(cs_cord, as_vtk_matrix=True)
            i_matrix.Invert()
        else:
            matrix = vtkMatrix4x4()
            i_matrix = vtkMatrix4x4()

        shp = (self.n_sector, result.shape[0], result.shape[1])
        full_result = np.empty(shp)
        full_result[:] = result

        rang = 360.0 / self.n_sector
        for i in range(1, self.n_sector):
            # transform to standard position, rotate about Z axis,
            # transform back
            transform = vtkTransform()
            transform.RotateZ(rang*i)
            transform.Update()
            rot_matrix = transform.GetMatrix()

            if cs_cord > 1:
                temp_matrix = vtkMatrix4x4()
                rot_matrix.Multiply4x4(i_matrix, rot_matrix, temp_matrix)
                rot_matrix.Multiply4x4(temp_matrix, matrix, rot_matrix)

            trans = pv.trans_from_matrix(rot_matrix)
            if tensor:
                if stress:
                    _binary_reader.tensor_arbitrary(full_result[i], trans)
                else:
                    _binary_reader.tensor_strain_arbitrary(full_result[i], trans)
            else:
                _binary_reader.affline_transform(full_result[i], trans)

        return full_result

    def _expand_cyclic_modal(self, result, result_r, hindex, phase, as_complex,
                             full_rotor):
        """ Combines repeated results from ANSYS """
        if as_complex or full_rotor:
            # Matches ansys direction
            # if self._positive_cyclic_dir:
            result_combined = result + result_r*1j
            # else:
                # result_combined = result - result_r*1j
            if phase:
                result_combined *= 1*np.cos(phase) - 1j*np.sin(phase)
        else:  # convert to real
            result_combined = result*np.cos(phase) - result_r*np.sin(phase)

        # just return single sector
        if not full_rotor:
            return result_combined

        # Generate full rotor solution
        result_expanded = []
        angles = np.linspace(0, 2*np.pi, self.n_sector + 1)[:-1] + phase
        for angle in angles:
            # need to rotate solution and rotate direction
            result_expanded.append(axis_rotation(result_combined,
                                                 angle, deg=False,
                                                 axis='z'))

        result_expanded = np.asarray(result_expanded)

        # ANSYS scales the result
        if hindex == 0 or hindex == self.n_sector/2:
            result_expanded /= self.n_sector**0.5
        else:
            result_expanded /= (self.n_sector/2)**0.5

        # adjust phase of the full result based on the harmonic index
        f_arr = np.zeros(self.n_sector)
        f_arr[hindex] = 1
        jang = np.fft.ifft(f_arr)[:self.n_sector]*self.n_sector
        cjang = jang * (np.cos(phase) - np.sin(phase) * 1j)  # 14-233

        result_expanded *= cjang.reshape(-1, 1, 1)
        if as_complex:
            return result_expanded
        else:
            return np.real(result_expanded)

    def _expand_cyclic_modal_tensor(self, result, result_r, hindex,
                                    phase, as_complex, full_rotor, stress=True):
        """Combines repeated results from ANSYS and optionally
        duplicates/rotates it around the axis of rotation"""

        # must scale value

        if as_complex or full_rotor:
            # if self._positive_cyclic_dir:
            result_combined = result + result_r*1j
            # else:
                # result_combined = result - result_r*1j

            if phase:
                result_combined *= 1*np.cos(phase) - 1j*np.sin(phase)

        else:  # convert to real
            result_combined = result*np.cos(phase) - result_r*np.sin(phase)

        # just return single sector
        if not full_rotor:
            return result_combined

        # Generate full rotor solution
        shp = (self.n_sector, result.shape[0], result.shape[1])
        result_expanded = np.empty(shp,np.complex128)
        result_expanded[:] = result_combined

        # convert hindex to nodal content
        f_arr = np.zeros(self.n_sector)
        f_arr[hindex] = self.n_sector

        jang = np.fft.ifft(f_arr)[:self.n_sector]

        # must be adjusted of to expand
        if hindex == 0 or hindex == self.n_sector/2:
            jang *= self.n_sector**(-1/2)
        else:
            jang *= 2*(2*self.n_sector)**(-1/2)

        cjang = jang * (np.cos(phase) - np.sin(phase) * 1j)

        full_result = np.empty(shp)
        full_result[:] = np.real(result_expanded*cjang.reshape(-1, 1, 1))

        cs_cord = self._resultheader['csCord']
        if cs_cord > 1:
            matrix = self.cs_4x4(cs_cord, as_vtk_matrix=True)
            i_matrix = self.cs_4x4(cs_cord, as_vtk_matrix=True)
            i_matrix.Invert()
        else:
            matrix = vtkMatrix4x4()
            i_matrix = vtkMatrix4x4()

        rang = 360.0 / self.n_sector
        for i in range(1, self.n_sector):
            # transform to standard position, rotate about Z axis,
            # transform back
            transform = vtkTransform()
            transform.RotateZ(rang*i)
            transform.Update()
            rot_matrix = transform.GetMatrix()

            if cs_cord > 1:
                temp_matrix = vtkMatrix4x4()
                rot_matrix.Multiply4x4(i_matrix, rot_matrix, temp_matrix)
                rot_matrix.Multiply4x4(temp_matrix, matrix, rot_matrix)

            trans = pv.trans_from_matrix(rot_matrix)
            if stress:
                _binary_reader.tensor_arbitrary(full_result[i], trans)
            else:
                _binary_reader.tensor_strain_arbitrary(full_result[i], trans)

        return full_result

    def harmonic_index_to_cumulative(self, hindex, mode):
        """Converts a harmonic index and a 0 index mode number to a
        cumulative result index.

        Harmonic indices are stored as positive and negative pairs for
        modes other than 0 and N/nsectors.

        Parameters
        ----------
        hindex : int
            Harmonic index.  Must be less than or equal to nsectors/2.
            May be positive or negative

        mode : int
            Mode number.  0 based indexing.  Access mode pairs by with
            a negative/positive harmonic index.

        Returns
        -------
        rnum : int
            Cumulative index number.  Zero based indexing.

        """
        hindex_table = self._resultheader['hindex']
        if not np.any(abs(hindex) == np.abs(hindex_table)):
            raise Exception('Invalid harmonic index.\n' +
                            'Available indices: %s' % np.unique(hindex_table))

        mask = np.logical_and(hindex == hindex_table,
                              mode == self.mode_table)

        if not mask.any():
            mode_mask = abs(hindex) == np.abs(hindex_table)
            avail_modes = np.unique(self.mode_table[mode_mask])
            raise Exception('Invalid mode for harmonic index %d\n' % hindex +
                            'Available modes: %s' % avail_modes)

        index = mask.nonzero()[0]
        assert index.size == 1, 'Multiple cumulative index matches'
        return index[0]

    @property
    def mode_table(self):
        """Unique modes for cyclic results"""
        hindex_table = self._resultheader['hindex']
        diff = np.diff(np.abs(hindex_table))
        freqs = self.time_values
        mode_table = [0]
        c = 0
        for i in range(1, freqs.size):
            if diff[i - 1]:
                c = 0
                mode_table.append(c)
            elif np.isclose(freqs[i], freqs[i - 1]):
                mode_table.append(c)
            else:
                c += 1
                mode_table.append(c)
        return np.asarray(mode_table)

    @property
    def harmonic_indices(self):
        """Harmonic indices of the result file.

        Harmonic index is simply the Nodal Diameter of the mode.  This
        is defined as the number of complete sine waves that pass
        through the circumference.

        Examples
        --------
        >>> rst = pyansys.read_binary('file.rst')
        >>> rst.harmonic_indices
        array([ 0,  0,  0,  0,  0,  0, -1,  1, -1,  1,  1, -1,
               -2,  2, -2,  2, -2,  2,  3,  3,  3,  3,  3,  3], dtype=int32)
        """
        return self._resultheader['hindex']

    def nodal_stress(self, rnum, phase=0, as_complex=False, full_rotor=False):
        """Retrieves the component stresses for each node in the
        solution.

        The order of the results corresponds to the sorted node
        numbering.

        Computes the nodal stress by averaging the stress for each
        element at each node.  Due to the discontinuities across
        elements, stresses will vary based on the element they are
        evaluated from.

        Parameters
        ----------
        rnum : int or list
            Cumulative result number with zero based indexing, or a
            list containing (step, substep) of the requested result.

        phase : float
            Phase adjustment of the stress in degrees.

        as_complex : bool, optional
            Reports stress as a complex result.  Real and imaginary
            stresses correspond to the stress of the main and repeated
            sector.  Stress can be "rotated" using the phase
            parameter.

        full_rotor : bool, optional
            Expands the results to the full rotor when True.  Default
            False.

        Returns
        -------
        nodenum : numpy.ndarray
            Node numbers of the result.

        stress : numpy.ndarray
            Stresses at Sx Sy Sz Sxy Syz Sxz averaged at each corner
            node.  For the corresponding node numbers, see where
            result is the result object.

        Examples
        --------
        >>> import pyansys
        >>> rst = pyansys.read_binary('file.rst')
        >>> nnum, stress = rst.nodal_stress(0)

        Notes
        -----
        Nodes without a stress value will be NAN.

        """
        func = super().nodal_stress
        return self._get_full_result(rnum, func, phase, full_rotor, as_complex,
                                     tensor=True, stress=True)

    def _get_full_result(self, rnum, func, phase, full_rotor, as_complex,
                         tensor=True, stress=False):
        """Return the full rotor result or the complex result for a cyclic model.

        rnum : int or list
            Cumulative result number with zero based indexing, or a
            list containing (step, substep) of the requested result.

        phase : float
            Phase adjustment of the stress in radians.

        tensor : bool
            True when the result is a stress/strain tensor.  False
            when a scalar or displacement value.

        stress : bool
            True when tensor is a stress.  False when tensor is a
            strain.  Ignored when not a tensor.

        """
        rnum = self.parse_step_substep(rnum)
        nnum, full_result = func(rnum)

        # full result may or may not contain the duplicate sector
        if self._has_duplicate_sector:
            result = full_result[self._mas_ind]
            nnum = nnum[self._mas_ind]
        else:
            result = full_result

        if self._resultheader['kan'] == 0:  # static result
            if full_rotor:
                expanded_result = self._expand_cyclic_static(result,
                                                             tensor=tensor,
                                                             stress=stress)
            else:
                return nnum, result
        elif self._resultheader['kan'] == 2:  # modal analysis
            # combine modal solution results
            hindex_table = self._resultheader['hindex']
            hindex = hindex_table[rnum]  # move this to expand_modal_tensor
            result_r = self._get_complex_result(func, rnum, result)
            if tensor:
                expanded_result = self._expand_cyclic_modal_tensor(result,
                                                                   result_r,
                                                                   hindex,
                                                                   phase,
                                                                   as_complex,
                                                                   full_rotor,
                                                                   stress=stress)
            else:
                assert result.shape == result_r.shape
                expanded_result = self._expand_cyclic_modal(result,
                                                            result_r,
                                                            hindex,
                                                            phase,
                                                            as_complex,
                                                            full_rotor)
        else:
            raise RuntimeError('Unsupported analysis type')

        return nnum, expanded_result

    def _get_complex_result(self, func, rnum, full_result):
        """Acquire the duplicate sector or repeated result.

        Depending on the version of MAPDL, this may mean using the
        result from the duplicate sector or the mode pair when there
        are duplicate modes.

        When there is no repeated mode or duplicate sector, returns an
        all zero array.

        Parameters
        ----------
        func : function
            Function to acquire the sector only result.

        rnum : int
            Cumulative result number.

        full_result : np.ndarray
            Full result (may include duplicate sector).

        Returns
        --------
        result_r : np.ndarray
            Repeated result
        """
        has_dup_result = False
        if self._has_duplicate_sector:
            has_dup_result = self._dup_ind[-1] <= full_result.shape[0] - 1

        if self._is_repeated_mode[rnum]:  # get mode pair result
            _, result_r = func(self._repeated_index[rnum])
            if result_r.shape[0] != self._mas_ind.size:
                result_r = result_r[self._mas_ind]
        elif has_dup_result:  # use the duplicate sector
            result_r = full_result[self._dup_ind]
        else:  # otherwise, a standing wave (no complex component)            
            result_r = np.zeros((self._mas_ind.size, full_result.shape[1]),
                                dtype=full_result.dtype)

        return result_r

    def nodal_thermal_strain(self, rnum, phase=0, as_complex=False, full_rotor=False):
        """Nodal component thermal strains.  This record contains
        strains in the order X, Y, Z, XY, YZ, XZ, EQV, and eswell
        (element swelling strain).  Thermal strains are always values
        at the integration points moved to the nodes.

        Parameters
        ----------
        rnum : int or list
            Cumulative result number with zero based indexing, or a
            list containing (step, substep) of the requested result.

        phase : float
            Phase adjustment of the stress in degrees.

        as_complex : bool, optional
            Reports stress as a complex result.  Real and imaginary
            stresses correspond to the stress of the main and repeated
            sector.  Stress can be "rotated" using the phase
            parameter.

        full_rotor : bool, optional
            Expands the results to the full rotor when True.  Default
            False.

        Returns
        -------
        nodenum : numpy.ndarray
            Node numbers of the result.

        thermal_strain : np.ndarray
            Nodal component plastic strains.  Array is in the order
            X, Y, Z, XY, YZ, XZ, EQV, ESWELL

        Examples
        --------
        Load the nodal thermal strain for the first result.

        >>> import pyansys
        >>> rst = pyansys.read_binary('file.rst')
        >>> nnum, thermal_strain = rst.nodal_thermal_strain(0)

        Notes
        -----
        Nodes without a strain will be NAN.

        """
        func = super().nodal_thermal_strain
        return self._get_full_result(rnum, func, phase, full_rotor, as_complex,
                                     tensor=True, stress=False)

    def plot_nodal_thermal_strain(self, rnum,
                                  comp=None,
                                  phase=0,
                                  full_rotor=True,
                                  show_displacement=False,
                                  displacement_factor=1,
                                  node_components=None,
                                  element_components=None,
                                  sel_type_all=True,
                                  add_text=True,
                                  overlay_wireframe=False,
                                  **kwargs):
        """Plot nodal thermal strain.

        Parameters
        ----------
        rnum : int or list
            Cumulative result number with zero based indexing, or a
            list containing (step, substep) of the requested result.

        comp : str, optional
            Thermal strain component to display.  Available options:
            - ``"X"``
            - ``"Y"``
            - ``"Z"``
            - ``"XY"``
            - ``"YZ"``
            - ``"XZ"``
            - ``"EQV"``
            - ``"ESWELL"`` (element swelling strain)

        phase : float, optional
            Phase angle of the modal result in radians.  Only valid
            when full_rotor is True.  Default 0.

        full_rotor : bool, optional
            Expand the sector solution to the full rotor.

        show_displacement : bool, optional
            Deforms mesh according to the result.

        displacement_factor : float, optional
            Increases or decreases displacement by a factor.

        node_components : list, optional
            Accepts either a string or a list strings of node
            components to plot.  For example:
            ``['MY_COMPONENT', 'MY_OTHER_COMPONENT]``

        element_components : list, optional
            Accepts either a string or a list strings of element
            components to plot.  For example:
            ``['MY_COMPONENT', 'MY_OTHER_COMPONENT]``

        sel_type_all : bool, optional
            If node_components is specified, plots those elements
            containing all nodes of the component.  Default True.

        Returns
        -------
        cpos : list
            Camera position from vtk render window.

        Examples
        --------
        Plot nodal thermal strain for an academic rotor

        >>> import pyansys
        >>> result = pyansys.download_academic_rotor()
        >>> result.plot_nodal_thermal_strain(0)

        """
        if not full_rotor:
            return super().plot_nodal_thermal_strain(rnum, **kwargs)

        idx = check_comp(THERMAL_STRAIN_TYPES, comp)
        _, strain = self.nodal_thermal_strain(rnum, phase, False, True)
        scalars = strain[:, :, idx]

        kwargs.setdefault('stitle', '%s Nodal Thermal Strain' % comp)
        kwargs['node_components'] = node_components
        kwargs['element_components'] = element_components
        kwargs['show_displacement'] = show_displacement
        kwargs['displacement_factor'] = displacement_factor
        kwargs['overlay_wireframe'] = overlay_wireframe
        kwargs['add_text'] = add_text
        kwargs['node_components'] = node_components
        kwargs['element_components'] = element_components
        kwargs['sel_type_all'] = sel_type_all
        kwargs['phase'] = phase
        return self._plot_cyclic_point_scalars(scalars, rnum, **kwargs)

    def nodal_elastic_strain(self, rnum, phase=0, as_complex=False,
                             full_rotor=False):
        """Nodal component elastic strains.  This record contains
        strains in the order X, Y, Z, XY, YZ, XZ, EQV.

        Elastic strains can be can be nodal values extrapolated from
        the integration points or values at the integration points
        moved to the nodes.

        Parameters
        ----------
        rnum : int or list
            Cumulative result number with zero based indexing, or a
            list containing (step, substep) of the requested result.

        phase : float
            Phase adjustment of the stress in radians.

        as_complex : bool, optional
            Reports stress as a complex result.  Real and imaginary
            stresses correspond to the stress of the main and repeated
            sector.  Stress can be "rotated" using the phase
            parameter.

        full_rotor : bool, optional
            Expands the results to the full rotor when True.  Default
            False.

        Returns
        -------
        nodenum : numpy.ndarray
            Node numbers of the result.

        elastic_strain : numpy.ndarray
            Nodal component elastic strains.  Array is in the order
            X, Y, Z, XY, YZ, XZ, EQV.

        Examples
        --------
        Load the nodal elastic strain for the first result.

        >>> import pyansys
        >>> rst = pyansys.read_binary('file.rst')
        >>> nnum, elastic_strain = rst.nodal_stress(0)

        Notes
        -----
        Nodes without a strain will be NAN.

        """
        func = super().nodal_elastic_strain
        return self._get_full_result(rnum, func, phase, full_rotor, as_complex,
                                     tensor=True, stress=False)

    def plot_nodal_elastic_strain(self, rnum,
                                  comp=None, phase=0,
                                  full_rotor=True,
                                  show_displacement=False,
                                  displacement_factor=1,
                                  node_components=None,
                                  element_components=None,
                                  sel_type_all=True,
                                  add_text=True,
                                  overlay_wireframe=False,
                                  **kwargs):
        """Plot nodal elastic strain.

        Parameters
        ----------
        rnum : int or list
            Cumulative result number with zero based indexing, or a
            list containing (step, substep) of the requested result.

        comp : str, optional
            Elastic strain component to display.  Available options:
            - ``"X"``
            - ``"Y"``
            - ``"Z"``
            - ``"XY"``
            - ``"YZ"``
            - ``"XZ"``

        phase : float, optional
            Phase angle of the modal result in radians.  Only valid
            when full_rotor is True.  Default 0

        full_rotor : bool, optional
            Expand the sector solution to the full rotor.

        show_displacement : bool, optional
            Deforms mesh according to the result.

        displacement_factor : float, optional
            Increases or decreases displacement by a factor.

        node_components : list, optional
            Accepts either a string or a list strings of node
            components to plot.  For example:
            ``['MY_COMPONENT', 'MY_OTHER_COMPONENT]``

        element_components : list, optional
            Accepts either a string or a list strings of element
            components to plot.  For example:
            ``['MY_COMPONENT', 'MY_OTHER_COMPONENT]``

        sel_type_all : bool, optional
            If node_components is specified, plots those elements
            containing all nodes of the component.  Default True.

        Returns
        -------
        cpos : list
            Camera position from vtk render window.

        Examples
        --------
        Plot nodal elastic strain for an academic rotor

        >>> import pyansys
        >>> result = pyansys.download_academic_rotor()
        >>> result.plot_nodal_elastic_strain(0, 'X')

        """
        idx = check_comp(STRAIN_TYPES[:-1], comp)
        _, strain = self.nodal_elastic_strain(rnum, phase, False, full_rotor)
        scalars = strain[:, :, idx]

        kwargs.setdefault('stitle', '%s Nodal Elastic Strain' % comp.upper())
        kwargs['node_components'] = node_components
        kwargs['element_components'] = element_components
        kwargs['show_displacement'] = show_displacement
        kwargs['displacement_factor'] = displacement_factor
        kwargs['overlay_wireframe'] = overlay_wireframe
        kwargs['add_text'] = add_text
        kwargs['node_components'] = node_components
        kwargs['element_components'] = element_components
        kwargs['sel_type_all'] = sel_type_all
        kwargs['phase'] = phase
        return self._plot_cyclic_point_scalars(scalars, rnum, **kwargs)

    def nodal_plastic_strain(self, rnum, phase=0, as_complex=False,
                             full_rotor=False):
        """Nodal component plastic strains.  This record contains
        strains in the order X, Y, Z, XY, YZ, XZ, EQV.

        Plastic strains can be can be nodal values extrapolated from
        the integration points or values at the integration points
        moved to the nodes.

        Parameters
        ----------
        rnum : int or list
            Cumulative result number with zero based indexing, or a
            list containing (step, substep) of the requested result.

        phase : float
            Phase adjustment of the stress in degrees.

        as_complex : bool, optional
            Reports stress as a complex result.  Real and imaginary
            stresses correspond to the stress of the main and repeated
            sector.  Stress can be "rotated" using the phase
            parameter.

        full_rotor : bool, optional
            Expands the results to the full rotor when True.  Default
            False.

        Returns
        -------
        nodenum : numpy.ndarray
            Node numbers of the result.

        plastic_strain : numpy.ndarray
            Nodal component plastic strains.  Array is in the order
            X, Y, Z, XY, YZ, XZ, EQV.

        Examples
        --------
        Load the nodal plastic strain for the first result.

        >>> import pyansys
        >>> rst = pyansys.read_binary('file.rst')
        >>> nnum, plastic_strain = rst.nodal_stress(0)

        Notes
        -----
        Nodes without a strain will be NAN.

        """
        func = super().nodal_plastic_strain
        return self._get_full_result(rnum, func, phase, full_rotor, as_complex,
                                     tensor=True, stress=False)

    def plot_nodal_plastic_strain(self, rnum,
                                  comp=None, phase=0,
                                  full_rotor=True,
                                  show_displacement=False,
                                  displacement_factor=1,
                                  node_components=None,
                                  element_components=None,
                                  sel_type_all=True,
                                  add_text=True,
                                  overlay_wireframe=False,
                                  **kwargs):
        """Plot nodal plastic strain.

        Parameters
        ----------
        rnum : int or list
            Cumulative result number with zero based indexing, or a
            list containing (step, substep) of the requested result.

        comp : str, optional
            Plastic strain component to display.  Available options:
            - ``"X"``
            - ``"Y"``
            - ``"Z"``
            - ``"XY"``
            - ``"YZ"``
            - ``"XZ"``

        phase : float, optional
            Phase angle of the modal result in radians.  Only valid
            when full_rotor is True.  Default 0

        full_rotor : bool, optional
            Expand the sector solution to the full rotor.

        show_displacement : bool, optional
            Deforms mesh according to the result.

        displacement_factor : float, optional
            Increases or decreases displacement by a factor.

        node_components : list, optional
            Accepts either a string or a list strings of node
            components to plot.  For example:
            ``['MY_COMPONENT', 'MY_OTHER_COMPONENT]``

        element_components : list, optional
            Accepts either a string or a list strings of element
            components to plot.  For example:
            ``['MY_COMPONENT', 'MY_OTHER_COMPONENT]``

        sel_type_all : bool, optional
            If node_components is specified, plots those elements
            containing all nodes of the component.  Default True.

        Returns
        -------
        cpos : list
            Camera position from vtk render window.

        Examples
        --------
        Plot nodal plastic strain for an academic rotor

        >>> import pyansys
        >>> result = pyansys.download_academic_rotor()
        >>> result.plot_nodal_plastic_strain(0)

        """
        idx = check_comp(STRAIN_TYPES[:-1], comp)
        _, strain = self.nodal_plastic_strain(rnum, phase, False, full_rotor)
        scalars = strain[:, :, idx]

        kwargs.setdefault('stitle', '%s Nodal Plastic Strain' % comp)
        kwargs['node_components'] = node_components
        kwargs['element_components'] = element_components
        kwargs['show_displacement'] = show_displacement
        kwargs['displacement_factor'] = displacement_factor
        kwargs['overlay_wireframe'] = overlay_wireframe
        kwargs['add_text'] = add_text
        kwargs['node_components'] = node_components
        kwargs['element_components'] = element_components
        kwargs['sel_type_all'] = sel_type_all
        kwargs['phase'] = phase
        return self._plot_cyclic_point_scalars(scalars, rnum, **kwargs)

    def principal_nodal_stress(self, rnum, phase=0, as_complex=False,
                               full_rotor=False):
        """Computes the principal component stresses for each node in
        the solution.

        Parameters
        ----------
        rnum : int or list
            Cumulative result number with zero based indexing, or a
            list containing (step, substep) of the requested result.

        phase : float
            Phase adjustment of the stress in degrees.

        as_complex : bool, optional
            Returns result as a complex number, otherwise as the real
            part rotated by phase.  Default False.

        full_rotor : bool, optional
            Expand sector solution to full rotor.

        Returns
        -------
        nodenum : numpy.ndarray
            Node numbers of the result.

        pstress : numpy.ndarray
            Principal stresses, stress intensity, and equivalant stress.
            [sigma1, sigma2, sigma3, sint, seqv]

        Notes
        -----
        ANSYS equivalant of:
        PRNSOL, S, PRIN

        which returns:
        S1, S2, S3 principal stresses, SINT stress intensity, and SEQV
        equivalent stress.
        """
        if as_complex and full_rotor:
            raise ValueError('complex and full_rotor cannot both be True')

        # get component stress
        nnum, stress = self.nodal_stress(rnum, phase, as_complex, full_rotor)

        # compute principle stress
        if as_complex:
            stress_r = np.imag(stress)
            stress = np.real(stress)

            pstress, isnan = _binary_reader.compute_principal_stress(stress)
            pstress[isnan] = np.nan
            pstress_r, isnan = _binary_reader.compute_principal_stress(stress_r)
            pstress_r[isnan] = np.nan

            return nnum, pstress + 1j*pstress_r

        elif full_rotor:
            # compute principle stress for each sector
            pstress = np.empty((self.n_sector, stress.shape[1], 5), np.float64)
            for i in range(stress.shape[0]):
                pstress[i], isnan = _binary_reader.compute_principal_stress(stress[i])
                pstress[i, isnan] = np.nan
            return nnum, pstress

        else:
            pstress, isnan = _binary_reader.compute_principal_stress(stress)
            pstress[isnan] = np.nan
            return nnum, pstress

    def plot_nodal_solution(self, rnum, comp='norm',
                            phase=0,
                            full_rotor=True,
                            show_displacement=False,
                            displacement_factor=1.0,
                            node_components=None,
                            element_components=None,
                            overlay_wireframe=False,
                            add_text=True,
                            sel_type_all=True,
                            **kwargs):
        """Plot the nodal solution (generally displacement).

        Parameters
        ----------
        rnum : int or list
            Cumulative result number with zero based indexing, or a
            list containing (step, substep) of the requested result.

        comp : str, optional
            Display component to display.  Options are 'x', 'y', 'z',
            and 'norm', corresponding to the x directin, y direction,
            z direction, and the normalized direction:
            ``(x**2 + y**2 + z**2)**0.5``

        full_rotor : bool, optional
            Expand sector solution to full rotor.

        phase : float, optional
            Phase angle of the modal result in radians.  Only valid
            when full_rotor is True.  Default 0

        node_components : list, optional
            Accepts either a string or a list strings of node
            components to plot.  For example:
            ``['MY_COMPONENT', 'MY_OTHER_COMPONENT]``

        element_components : list, optional
            Accepts either a string or a list strings of element
            components to plot.  For example:
            ``['MY_COMPONENT', 'MY_OTHER_COMPONENT]``

        sel_type_all : bool, optional
            If node_components is specified, plots those elements
            containing all nodes of the component.  Default True.

        Returns
        -------
        cpos : list
            Camera position from vtk render window.

        """

        # Load result from file
        if not full_rotor:
            return super().plot_nodal_solution(rnum,
                                               comp,
                                               show_displacement=show_displacement,
                                               displacement_factor=displacement_factor,
                                               node_components=node_components,
                                               element_components=element_components,
                                               sel_type_all=sel_type_all,
                                               **kwargs)

        rnum = self.parse_step_substep(rnum)
        _, result = self.nodal_solution(rnum, phase, full_rotor, as_complex=False)

        # Process result
        label = 'Cyclic Rotor\nDisplacement'
        if comp == 'x':
            scalars = result[:, :, 0]
            stitle = 'X {:s}\n'.format(label)
        elif comp == 'y':
            scalars = result[:, :, 1]
            stitle = 'Y {:s}\n'.format(label)
        elif comp == 'z':
            scalars = result[:, :, 2]
            stitle = 'Z {:s}\n'.format(label)
        else:
            # Normalize displacement
            scalars = (result*result).sum(2)**0.5
            stitle = 'Normalized\n%s\n' % label
        kwargs.setdefault('stitle', stitle)

        kwargs['node_components'] = node_components
        kwargs['element_components'] = element_components
        kwargs['show_displacement'] = show_displacement
        kwargs['displacement_factor'] = displacement_factor
        kwargs['overlay_wireframe'] = overlay_wireframe
        kwargs['add_text'] = add_text
        kwargs['node_components'] = node_components
        kwargs['element_components'] = element_components
        kwargs['sel_type_all'] = sel_type_all
        kwargs['phase'] = phase
        return self._plot_cyclic_point_scalars(scalars, rnum, **kwargs)

    def plot_nodal_stress(self, rnum,
                          comp=None,
                          phase=0,
                          full_rotor=True,
                          show_displacement=False,
                          displacement_factor=1,
                          node_components=None,
                          element_components=None,
                          overlay_wireframe=False,
                          add_text=True,
                          sel_type_all=True, **kwargs):
        """Plot nodal stress of a given component

        Parameters
        ----------
        rnum : int or list
            Cumulative result number with zero based indexing, or a
            list containing (step, substep) of the requested result.

        comp : str, optional
            Stress component to display.  Available options:
            - ``"X"``
            - ``"Y"``
            - ``"Z"``
            - ``"XY"``
            - ``"YZ"``
            - ``"XZ"``

        phase : float, optional
            Phase angle of the modal result in radians.  Only valid
            when full_rotor is True.  Default 0

        full_rotor : bool, optional
            Expand the sector solution to the full rotor.

        show_displacement : bool, optional
            Deforms mesh according to the result.

        displacement_factor : float, optional
            Increases or decreases displacement by a factor.

        node_components : list, optional
            Accepts either a string or a list strings of node
            components to plot.  For example:
            ``['MY_COMPONENT', 'MY_OTHER_COMPONENT]``

        element_components : list, optional
            Accepts either a string or a list strings of element
            components to plot.  For example:
            ``['MY_COMPONENT', 'MY_OTHER_COMPONENT]``

        sel_type_all : bool, optional
            If node_components is specified, plots those elements
            containing all nodes of the component.  Default True.

        Returns
        -------
        cpos : list
            Camera position from vtk render window.

        """
        if not full_rotor:
            super().plot_nodal_stress(rnum, comp,
                                      show_displacement,
                                      displacement_factor,
                                      node_components,
                                      element_components,
                                      sel_type_all, **kwargs)

        idx = check_comp(STRESS_TYPES, comp)
        _, stress = self.nodal_stress(rnum, phase, False, full_rotor=True)
        scalars = stress[:, :, idx]

        kwargs.setdefault('stitle', 'Cyclic Rotor\nNodal Stress\n%s\n' % comp)
        kwargs['node_components'] = node_components
        kwargs['element_components'] = element_components
        kwargs['show_displacement'] = show_displacement
        kwargs['displacement_factor'] = displacement_factor
        kwargs['overlay_wireframe'] = overlay_wireframe
        kwargs['add_text'] = add_text
        kwargs['node_components'] = node_components
        kwargs['element_components'] = element_components
        kwargs['sel_type_all'] = sel_type_all
        kwargs['phase'] = phase
        return self._plot_cyclic_point_scalars(scalars, rnum, **kwargs)

    def plot_principal_nodal_stress(self, rnum,
                                    comp=None,
                                    phase=0,
                                    full_rotor=True,
                                    show_displacement=False,
                                    displacement_factor=1,
                                    node_components=None,
                                    element_components=None,
                                    sel_type_all=True,
                                    add_text=True,
                                    overlay_wireframe=False,
                                    **kwargs):
        """Plot the nodal principal stress.

        Parameters
        ----------
        rnum : int or list
            Cumulative result number with zero based indexing, or a
            list containing (step, substep) of the requested result.

        comp : string
            Stress component to plot.  S1, S2, S3 principal stresses, SINT
            stress intensity, and SEQV equivalent stress.

            Stress type must be a string from the following list:
            ``['S1', 'S2', 'S3', 'SINT', 'SEQV']``

        phase : float, optional
            Phase angle of the modal result in radians.  Only valid
            when full_rotor is True.  Default 0

        full_rotor : bool, optional
            Expand sector solution to full rotor.

        show_displacement : bool, optional
            Deforms mesh according to the result.

        displacement_factor : float, optional
            Increases or decreases displacement by a factor.

        node_components : list, optional
            Accepts either a string or a list strings of node
            components to plot.  For example:
            ``['MY_COMPONENT', 'MY_OTHER_COMPONENT]``

        sel_type_all : bool, optional
            If node_components is specified, plots those elements
            containing all nodes of the component.  Default True.

        kwargs : keyword arguments
            Additional keyword arguments.  See ``help(pyvista.plot)``

        Returns
        -------
        cpos : list
            VTK camera position.

        """
        if not full_rotor:
            return super().plot_principal_nodal_stress(rnum, comp,
                                                       show_displacement,
                                                       displacement_factor,
                                                       node_components,
                                                       element_components,
                                                       sel_type_all, **kwargs)

        # get the correct component of the principal stress for the rotor
        idx = check_comp(PRINCIPAL_STRESS_TYPES, comp)
        _, pstress = self.principal_nodal_stress(rnum, phase, full_rotor=True)
        scalars = pstress[:, :, idx]

        kwargs.setdefault('stitle',
                          'Cyclic Rotor\nPrincipal Nodal Stress\n%s\n' % comp)
        kwargs['node_components'] = node_components
        kwargs['element_components'] = element_components
        kwargs['show_displacement'] = show_displacement
        kwargs['displacement_factor'] = displacement_factor
        kwargs['overlay_wireframe'] = overlay_wireframe
        kwargs['add_text'] = add_text
        kwargs['node_components'] = node_components
        kwargs['element_components'] = element_components
        kwargs['sel_type_all'] = sel_type_all
        kwargs['phase'] = phase
        self._plot_cyclic_point_scalars(scalars, rnum, **kwargs)

    def nodal_temperature(self, rnum, full_rotor=False):
        """Retrieves the temperature for each node in the
        solution.

        The order of the results corresponds to the sorted node
        numbering.

        Equivalent MAPDL commands:
        PRNSOL, TEMP
        PRNSOL, BFE

        Parameters
        ----------
        rnum : int or list
            Cumulative result number with zero based indexing, or a
            list containing (step, substep) of the requested result.

        full_rotor : bool, optional
            Expand sector solution to full rotor.

        Returns
        -------
        nnum : numpy.ndarray
            Node numbers of the result.

        temperature : numpy.ndarray
            Temperature at each node.

        Examples
        --------
        >>> import pyansys
        >>> rst = pyansys.read_binary('file.rst')
        >>> nnum, stress = rst.nodal_temperature(0)

        """
        nnum, temp = super()._nodal_result(rnum, 'EPT')
        nnum = nnum[self._mas_ind]
        temp = temp[self._mas_ind]
        if not full_rotor:  # return only the master sector
            return nnum, temp.ravel()

        # otherwise, duplicate and repeat as temperature is constant across sectors
        return nnum, temp.T.repeat(self.n_sector, axis=0)

    def plot_nodal_temperature(self, rnum,
                               phase=0,
                               full_rotor=True,
                               show_displacement=False,
                               displacement_factor=1.0,
                               node_components=None,
                               overlay_wireframe=False,
                               add_text=True,
                               element_components=None,
                               sel_type_all=True,
                               **kwargs):
        """Plot the nodal temperature.

        Parameters
        ----------
        rnum : int or list
            Cumulative result number with zero based indexing, or a
            list containing (step, substep) of the requested result.

        full_rotor : bool, optional
            Expand the sector solution and plot the full rotor.

        phase : float, optional
            Phase angle of the modal result in radians.  Only valid
            when full_rotor is True.  Default 0

        node_components : list, optional
            Accepts either a string or a list strings of node
            components to plot.  For example:
            ``['MY_COMPONENT', 'MY_OTHER_COMPONENT]``

        element_components : list, optional
            Accepts either a string or a list strings of element
            components to plot.  For example:
            ``['MY_COMPONENT', 'MY_OTHER_COMPONENT]``

        sel_type_all : bool, optional
            If node_components is specified, plots those elements
            containing all nodes of the component.  Default True.

        Returns
        -------
        cpos : list
            Camera position from vtk render window.

        """
        # Load result from file
        if not full_rotor:
            return super().plot_nodal_temperature(rnum,
                                                  show_displacement=show_displacement,
                                                  displacement_factor=displacement_factor,
                                                  node_components=node_components,
                                                  element_components=element_components,
                                                  sel_type_all=sel_type_all,
                                                  **kwargs)

        _, temp = self.nodal_temperature(rnum, True)

        kwargs['node_components'] = node_components
        kwargs['element_components'] = element_components
        kwargs['show_displacement'] = show_displacement
        kwargs['displacement_factor'] = displacement_factor
        kwargs['overlay_wireframe'] = overlay_wireframe
        kwargs['add_text'] = add_text
        kwargs['node_components'] = node_components
        kwargs['element_components'] = element_components
        kwargs['sel_type_all'] = sel_type_all
        kwargs['phase'] = phase
        return self._plot_cyclic_point_scalars(temp, rnum, **kwargs)

    def animate_nodal_solution(self, rnum, comp='norm',
                               displacement_factor=0.1,
                               nangles=180,
                               add_text=True, loop=True,
                               movie_filename=None,
                               **kwargs):
        """Animate nodal solution.  Assumes nodal solution is a
        displacement array from a modal solution.

        rnum : int or list
            Cumulative result number with zero based indexing, or a
            list containing (step, substep) of the requested result.

        comp : str, optional
            Component of displacement to display.  Options are 'x',
            'y', 'z', or 'norm', which correspond to the x , y, z, or
            the normalized direction ``(x**2 + y**2 + z**2)**0.5``

        displacement_factor : float, optional
            Increases or decreases displacement by a factor.

        nangles : int, optional
            Number of "frames" between each full cycle.

        show_phase : bool, optional
            Shows the phase at each frame.

        add_text : bool, optional
            Includes result information at the bottom left-hand corner
            of the plot.

        interpolate_before_map : bool, optional
            Leaving this at default generally results in a better plot.

        movie_filename : str, optional
            Filename of the movie to open.  Filename should end in mp4,
            but other filetypes may be supported.  See ``imagio.get_writer``.
            A single loop of the mode will be recorded.

        kwargs : optional keyword arguments, optional
            See help(pyvista.plot) for additional keyword arguments.

        """
        rnum = self.parse_step_substep(rnum)  # need cumulative
        if 'full_rotor' in kwargs:
            raise NotImplementedError('``full_rotor`` keyword argument not supported')

        # normalize nodal solution
        _, complex_disp = self.nodal_solution(rnum, as_complex=True,
                                              full_rotor=True)
        complex_disp *= displacement_factor
        complex_disp = complex_disp.reshape(-1, 3)

        if comp == 'x':
            axis = 0
        elif comp == 'y':
            axis = 1
        elif comp == 'z':
            axis = 2
        else:
            axis = None

        result_info = ''
        if add_text:
            result_info = self.text_result_table(rnum)

        # need only the surface of the full rotor
        plot_mesh = self.full_rotor.extract_surface()
        orig_pt = plot_mesh.points.copy()

        # reduce the complex displacement to just the surface points
        ind = plot_mesh.point_arrays['vtkOriginalPointIds']
        complex_disp = np.take(complex_disp, ind, axis=0)

        if axis is not None:
            scalars = complex_disp[:, axis]
        else:
            scalars = (complex_disp*complex_disp).sum(1)**0.5

        # intialize plotter
        text_color = kwargs.pop('text_color', None)
        cpos = kwargs.pop('cpos', None)
        off_screen = kwargs.pop('off_screen', None)
        plotter = pv.Plotter(off_screen=off_screen)
        if kwargs.pop('show_axes', True):
            plotter.add_axes()

        if 'rng' not in kwargs:
            smax = np.abs(scalars).max()
            if comp == 'norm':
                kwargs['rng'] = [0, smax]
            else:
                kwargs['rng'] = [-smax, smax]

        background = kwargs.pop('background', None)
        if background:
            plotter.set_background(background)

        plotter.add_mesh(plot_mesh,
                         scalars=np.real(scalars),
                         **kwargs)

        # setup text
        plotter.add_text(' ', font_size=20, position=[0, 0], color=text_color)

        if cpos:
            plotter.camera_position = cpos

        if movie_filename:
            if movie_filename.strip()[-3:] == 'gif':
                plotter.open_gif(movie_filename)
            else:
                plotter.open_movie(movie_filename)

        self._animating = True
        def q_callback():
            """exit when user wants to leave"""
            self._animating = False

        plotter.add_key_event("q", q_callback)

        # run until q is pressed
        plotter.show(interactive=False, auto_close=False,
                     interactive_update=not off_screen)

        first_loop = True
        while self._animating:
            for angle in np.linspace(0, np.pi*2, nangles):
                padj = 1*np.cos(angle) - 1j*np.sin(angle)
                complex_disp_adj = np.real(complex_disp*padj)

                if axis is not None:
                    scalars = complex_disp_adj[:, axis]
                else:
                    scalars = (complex_disp_adj*complex_disp_adj).sum(1)**0.5

                plotter.update_scalars(scalars, render=False)
                plot_mesh.points[:] = orig_pt + complex_disp_adj

                if add_text:
                    plotter.textActor.SetInput('%s\nPhase %.1f Degrees' %
                                               (result_info, (angle*180/np.pi)))

                plotter.update(1, force_redraw=True)
                if not self._animating:
                    break

                if movie_filename and first_loop:
                    plotter.write_frame()

            first_loop = False
            if not loop:
                break

        cpos = plotter.camera_position
        plotter.close()
        return cpos

    @wraps(animate_nodal_solution)
    def animate_nodal_displacement(self, *args, **kwargs):
        """wraps animate_nodal_solution"""
        return self.animate_nodal_solution(*args, **kwargs)

    def _gen_full_rotor(self):
        """ Create full rotor vtk unstructured grid """
        grid = self._mas_grid.copy()
        # transform to standard coordinate system
        cs_cord = self._resultheader['csCord']
        if cs_cord > 1:
            matrix = self.cs_4x4(cs_cord, as_vtk_matrix=True)
            grid.transform(matrix)

        # consider forcing low and high to be exact
        # self._mas_grid.point_arrays['CYCLIC_M01H'] --> rotate and match

        vtkappend = vtkAppendFilter()
        # vtkappend.MergePointsOn()
        # vtkappend.SetTolerance(1E-3)  # not available until vtk 9+
        rang = 360.0 / self.n_sector
        for i in range(self.n_sector):
            # Transform mesh
            sector = grid.copy()
            sector_id = np.empty(grid.n_points)
            sector_id[:] = i
            sector.point_arrays['sector_id'] = sector_id
            sector.rotate_z(rang * i)
            vtkappend.AddInputData(sector)

        vtkappend.Update()
        full_rotor = pv.wrap(vtkappend.GetOutput())

        if cs_cord > 1:
            matrix.Invert()
            full_rotor.transform(matrix)

        return full_rotor

    @property
    def full_rotor(self):
        """UnstructuredGrid of the full replicated rotor"""
        if self._rotor_cache is None:
            self._rotor_cache = self._gen_full_rotor()
        return self._rotor_cache

    def _plot_cyclic_point_scalars(self, scalars, rnum,
                                   show_displacement=False,
                                   displacement_factor=1,
                                   overlay_wireframe=False,
                                   add_text=True,
                                   node_components=None,
                                   element_components=None,
                                   sel_type_all=True,
                                   phase=None,
                                   **kwargs):
        """Plot point scalars on active mesh.

        Parameters
        ----------
        scalars : numpy.ndarray
            Node scalars to plot.

        rnum : int, optional
            Cumulative result number.  Used for adding informative
            text.

        grid : pyvista.PolyData or pyvista.UnstructuredGrid, optional
            Uses self.grid by default.  When specified, uses this grid
            instead.

        show_displacement : bool, optional
            Deforms mesh according to the result.

        displacement_factor : float, optional
            Increases or decreases displacement by a factor.

        overlay_wireframe : bool, optional
            Overlay a wireframe of the original undeformed mesh.

        add_text : bool, optional
            Adds information about the result when rnum is given.

        kwargs : keyword arguments
            Additional keyword arguments.  See ``help(pyvista.plot)``

        Returns
        -------
        cpos : list
            Camera position.

        """
        # extract a sub-component of the grid if requested
        grid = self._mas_grid
        if node_components:
            grid, ind = self._extract_node_components(node_components,
                                                      sel_type_all)
            if scalars is not None:
                scalars = scalars[:, ind]

        elif element_components:
            grid, ind = self._extract_element_components(element_components)
            if scalars is not None:
                scalars = scalars[:, ind]

        # must be removed before add_mesh **kwargs
        window_size = kwargs.pop('window_size', None)
        full_screen = kwargs.pop('full_screen', False)
        screenshot = kwargs.pop('screenshot', None)
        text_color = kwargs.pop('text_color', None)
        kwargs.setdefault('cmap', 'jet')
        if scalars is not None:
            kwargs.setdefault('rng', [np.nanmin(scalars), np.nanmax(scalars)])

        # Plot off screen when not interactive
        off_screen = kwargs.pop('off_screen', None)
        plotter = pv.Plotter(off_screen=off_screen)

        # various plotting properties that must be removed before add_mesh
        if kwargs.pop('show_axes', True):
            plotter.add_axes()
        plotter.background_color = kwargs.pop('background', None)
        cpos = kwargs.pop('cpos', None)

        cs_cord = self._resultheader['csCord']
        if cs_cord > 1:
            matrix = self.cs_4x4(cs_cord, as_vtk_matrix=True)
            i_matrix = self.cs_4x4(cs_cord, as_vtk_matrix=True)
            i_matrix.Invert()
        else:
            matrix = vtkMatrix4x4()
            i_matrix = vtkMatrix4x4()

        if overlay_wireframe:
            rang = 360.0 / self.n_sector
            for i in range(self.n_sector):
                actor = plotter.add_mesh(grid.copy(False), color='w',
                                         style='wireframe',
                                         opacity=0.5, **kwargs)

                # transform to standard position, rotate about Z axis,
                # transform back
                transform = vtkTransform()
                transform.RotateZ(rang*i)
                transform.Update()
                rot_matrix = transform.GetMatrix()

                if cs_cord > 1:
                    temp_matrix = vtkMatrix4x4()
                    rot_matrix.Multiply4x4(i_matrix, rot_matrix, temp_matrix)
                    rot_matrix.Multiply4x4(temp_matrix, matrix, rot_matrix)
                    transform.SetMatrix(rot_matrix)

                actor.SetUserTransform(transform)

        # add main mesh
        if show_displacement:
            _, disp = self.nodal_solution(rnum, phase, full_rotor=True)
            disp *= displacement_factor
            if node_components:
                _, ind = self._extract_node_components(node_components, sel_type_all)
                disp = disp[:, ind]
            elif element_components:
                _, ind = self._extract_element_components(element_components)
                disp = disp[:, ind]

            disp = disp.reshape(-1, 3)

            rotor = self.full_rotor.copy()
            rotor.points += disp
            actor = plotter.add_mesh(rotor,
                                     scalars=scalars.reshape(-1, 3),
                                     **kwargs)

        else:
            surf_sector = grid.extract_surface()
            ind = surf_sector.point_arrays['vtkOriginalPointIds']
            rang = 360.0 / self.n_sector
            for i in range(self.n_sector):
                if scalars is not None:
                    sector_scalars = scalars[i, ind]
                else:
                    sector_scalars = None

                actor = plotter.add_mesh(surf_sector.copy(False),
                                         scalars=sector_scalars,
                                         **kwargs)

                # NAN/missing data are white
                plotter.mapper.GetLookupTable().SetNanColor(1, 1, 1, 1)

                # transform to standard position, rotate about Z axis,
                # transform back
                transform = vtkTransform()
                transform.RotateZ(rang*i)
                transform.Update()
                rot_matrix = transform.GetMatrix()

                if cs_cord > 1:
                    temp_matrix = vtkMatrix4x4()
                    rot_matrix.Multiply4x4(i_matrix, rot_matrix, temp_matrix)
                    rot_matrix.Multiply4x4(temp_matrix, matrix, rot_matrix)
                    transform.SetMatrix(rot_matrix)

                actor.SetUserTransform(transform)

        # add table
        if isinstance(add_text, str):
            plotter.add_text(add_text, font_size=20, position=[0, 0],
                             color=text_color)
        elif add_text:
            rnum = self.parse_step_substep(rnum)
            plotter.add_text(self.text_result_table(rnum), font_size=20,
                             color=text_color)

        # must set camera position at the ended
        if cpos is not None:
            plotter.camera_position = cpos

        if screenshot:
            cpos = plotter.show(auto_close=False,
                                window_size=window_size,
                                full_screen=full_screen)
            plotter.screenshot(screenshot)
            plotter.close()
        else:
            cpos = plotter.show(window_size=window_size, full_screen=full_screen)

        return cpos



"""Contains the available result class

Based on resucm.inc included in the standard ANSYS unified install

Current as of v193 (which is the current result file as of v20.2

**********  AvailData bits  **********
 bit   data item

element data (see elparm.inc)
  1     EMS  - miscellaneous summable items(normally includes face pressures)
  2     ENF  - nodal forces
  3     ENS  - nodal stresses
  4     ENG  - element energies and volume
  5     EGR  - nodal gradients
  6     EEL  - nodal elastic strains
  7     EPL  - nodal plastic strains
  8     ECR  - nodal creep strains
  9     ETH  - nodal thermal strains (includes swelling strains)
 10     EUL  - element euler angles
 11     EFX  - nodal fluxes
 12     ELF  - nodal forces generated by the element, e.g. lorentz or maxwell forces
 13     EMN  - miscellaneous nonsummable items
 14     ECD  - nodal current densities
 15     ENL  - nodal nonlinear items, e.g. equivalent plastic strains
 16     EHC  - element joule heating
 17     EPT  - nodal temperatures
 18     SURF - face surface stresses
 19     EDI  - nodal diffusion strains
 20     ESTR - POST1 element (ETABLE) data
 21     ECT  - nodal contact items, e.g. penetration
 22     EXYZ - integration point locations
 23     EBAC - back stresses(for kinematic hardening)
 24     ESVR - saved variables from usermat
 25     EMR  - element material saved record
 26     ---

nodal data
 27     NSL  - Nodal displacements
 28     VSL  - Nodal velocities
 29     ASL  - Nodal accelerations
 30     RF   - Nodal reaction forces
 31     ---

"""


DISCRIPTION = {
    'EMS': 'miscellaneous summable items(normally includes face pressures)',
    'ENF': 'nodal forces',
    'ENS': 'nodal stresses',
    'ENG': 'element energies and volume',
    'EGR': 'nodal gradients',
    'EEL': 'nodal elastic strains',
    'EPL': 'nodal plastic strains',
    'ECR': 'nodal creep strains',
    'ETH': 'nodal thermal strains (includes swelling strains)',
    'EUL': 'element euler angles',
    'EFX': 'nodal fluxes',
    'ELF': 'nodal forces generated by the element, e.g. lorentz or maxwell forces',
    'EMN': 'miscellaneous nonsummable items',
    'ECD': 'nodal current densities',
    'ENL': 'nodal nonlinear items, e.g. equivalent plastic strains',
    'EHC': 'element joule heating',
    'EPT': 'nodal temperatures',
    'SURF': 'face surface stresses',
    'EDI': 'nodal diffusion strains',
    'ESTR': 'POST1 element (ETABLE) data',
    'ECT': 'nodal contact items, e.g. penetration',
    'EXYZ': 'integration point locations',
    'EBAC': 'back stresses(for kinematic hardening)',
    'ESVR': 'saved variables from usermat',
    # 'EMR': 'element material saved record',
    'NSL': 'Nodal displacements',
    'VSL': 'Nodal velocities',
    'ASL': 'Nodal accelerations',
    'RF': 'Nodal reaction forces',
}


class AvailableResults():

    def __init__(self, avail_bits, is_thermal):
        """Parse the available bits and determine if a given result is
        available.
        """
        self._avail_bits = avail_bits
        self._parsed_bits = {
            'EMS': self.ems,
            'ENF': self.enf,
            'ENS': self.ens,
            'ENG': self.eng,
            'EGR': self.egr,
            'EEL': self.eel,
            'EPL': self.epl,
            'ECR': self.ecr,
            'ETH': self.eth,
            'EUL': self.eul,
            'EFX': self.efx,
            'ELF': self.elf,
            'EMN': self.emn,
            'ECD': self.ecd,
            'ENL': self.enl,
            'EHC': self.ehc,
            'EPT': self.ept,
            'SURF': self.surf,
            'EDI': self.edi,
            'ESTR': self.estr,
            'ECT': self.ect,
            'EXYZ': self.exyz,
            'EBAC': self.ebac,
            'ESVR': self.esvr,
            # 'EMR': self.emr,
            'NSL': self.nsl,
            'VSL': self.vsl,
            'ASL': self.asl,
            'RF': self.rf,
            }

        self._valid_results = []
        for key, value in self._parsed_bits.items():
            if value:
                self._valid_results.append(key)

        self.description = DISCRIPTION
        if is_thermal:
            self.description['NSL'] = 'Nodal temperatures'

    def __getitem__(self, key):
        """Allow a key access"""
        return self._parsed_bits[key]

    def __iter__(self):
        for key in self._valid_results:
            yield key

    @property
    def ems(self):
        """Miscellaneous summable items(normally includes face pressures)"""
        return bool(self._avail_bits & 0b1 << 1)

    @property
    def enf(self):
        """Nodal forces"""
        return bool(self._avail_bits & 0b1 << 2)

    @property
    def ens(self):
        """Nodal stresses"""
        return bool(self._avail_bits & 0b1 << 3)

    @property
    def eng(self):
        """Element energies and volume"""
        return bool(self._avail_bits & 0b1 << 4)

    @property
    def egr(self):
        """Nodal gradients"""
        return bool(self._avail_bits & 0b1 << 5)

    @property
    def eel(self):
        """Nodal elastic strains"""
        return bool(self._avail_bits & 0b1 << 6)

    @property
    def epl(self):
        """Nodal plastic strains"""
        return bool(self._avail_bits & 0b1 << 7)

    @property
    def ecr(self):
        """Nodal creep strains"""
        return bool(self._avail_bits & 0b1 << 8)

    @property
    def eth(self):
        """Nodal thermal strains (includes swelling strains)"""
        return bool(self._avail_bits & 0b1 << 9)

    @property
    def eul(self):
        """Element euler angles"""
        return bool(self._avail_bits & 0b1 << 10)

    @property
    def efx(self):
        """Nodal fluxes"""
        return bool(self._avail_bits & 0b1 << 11)

    @property
    def elf(self):
        """Nodal forces generated by the element, e.g. lorentz or maxwell forces"""
        return bool(self._avail_bits & 0b1 << 12)

    @property
    def emn(self):
        """Miscellaneous nonsummable items"""
        return bool(self._avail_bits & 0b1 << 13)

    @property
    def ecd(self):
        """Nodal current densities"""
        return bool(self._avail_bits & 0b1 << 14)

    @property
    def enl(self):
        """Nodal nonlinear items, e.g. equivalent plastic strains"""
        return bool(self._avail_bits & 0b1 << 15)

    @property
    def ehc(self):
        """Element joule heating"""
        return bool(self._avail_bits & 0b1 << 16)

    @property
    def ept(self):
        """Nodal temperatures"""
        return bool(self._avail_bits & 0b1 << 17)

    @property
    def surf(self):
        """Face surface stresses"""
        return bool(self._avail_bits & 0b1 << 18)

    @property
    def edi(self):
        """Nodal diffusion strains"""
        return bool(self._avail_bits & 0b1 << 19)

    @property
    def estr(self):
        """Post1 element (ETABLE) data"""
        return bool(self._avail_bits & 0b1 << 20)

    @property
    def ect(self):
        """Nodal contact items, e.g. penetration"""
        return bool(self._avail_bits & 0b1 << 21)

    @property
    def exyz(self):
        """Integration point locations"""
        return bool(self._avail_bits & 0b1 << 22)

    @property
    def ebac(self):
        """Back stresses(for kinematic hardening)"""
        return bool(self._avail_bits & 0b1 << 23)

    @property
    def esvr(self):
        """Saved variables from usermat"""
        return bool(self._avail_bits & 0b1 << 24)

    # @property
    # def emr(self):
    #     """Element material saved record"""
    #     return bool(self._avail_bits & 0b1 << 25)

    @property
    def nsl(self):
        """Nodal displacements"""
        return bool(self._avail_bits & 0b1 << 27)

    @property
    def vsl(self):
        """Nodal velocities"""
        return bool(self._avail_bits & 0b1 << 28)

    @property
    def asl(self):
        """Nodal accelerations"""
        return bool(self._avail_bits & 0b1 << 29)

    @property
    def rf(self):
        """Nodal reaction forces"""
        return bool(self._avail_bits & 0b1 << 30)

    def __repr__(self):
        txt = "Available Results:\n"

        if self.ems:
            txt += "EMS : Miscellaneous summable items (normally includes face pressures)\n"
        if self.enf:
            txt += "ENF : Nodal forces\n"
        if self.ens:
            txt += "ENS : Nodal stresses\n"
        if self.eng:
            txt += "ENG : Element energies and volume\n"
        if self.egr:
            txt += "EGR : Nodal gradients\n"
        if self.eel:
            txt += "EEL : Nodal elastic strains\n"
        if self.epl:
            txt += "EPL : Nodal plastic strains\n"
        if self.ecr:
            txt += "ECR : Nodal creep strains\n"
        if self.eth:
            txt += "ETH : Nodal thermal strains (includes swelling strains)\n"
        if self.eul:
            txt += "EUL : Element euler angles\n"
        if self.efx:
            txt += "EFX : Nodal fluxes\n"
        if self.elf:
            txt += "ELF : Nodal forces generated by the element, e.g. lorentz or maxwell forces\n"
        if self.emn:
            txt += "EMN : Miscellaneous nonsummable items\n"
        if self.ecd:
            txt += "ECD : Nodal current densities\n"
        if self.enl:
            txt += "ENL : Nodal nonlinear items, e.g. equivalent plastic strains\n"
        if self.ehc:
            txt += "EHC : Element joule heating\n"
        if self.ept:
            txt += "EPT : Nodal temperatures\n"
        if self.surf:
            txt += "SURF: Face surface stresses\n"
        if self.edi:
            txt += "EDI : Nodal diffusion strains\n"
        if self.estr:
            txt += "ESTR: Post1 element (ETABLE) data\n"
        if self.ect:
            txt += "ECT : Nodal contact items, e.g. penetration\n"
        if self.exyz:
            txt += "EXYZ: Integration point locations\n"
        if self.ebac:
            txt += "EBAC: Back stresses (for kinematic hardening)\n"
        if self.esvr:
            txt += "ESVR: Saved variables from usermat\n"
        # if self.emr:
        #     txt += "EMR : Element material saved record\n"
        if self.nsl:
            txt += "NSL : Nodal displacements\n"
        if self.vsl:
            txt += "VSL : Nodal velocities\n"
        if self.asl:
            txt += "ASL : Nodal accelerations\n"
        if self.rf:
            txt += "RF  : Nodal reaction forces\n"

        return txt



import appdirs
import os

# Per contract with Sphinx-Gallery, this method must be available at top level
from pyvista.utilities.sphinx_gallery import _get_sg_image_scraper

from pyansys._version import __version__
from pyansys.archive import (Archive, write_cmblock, write_nblock,
                             save_as_archive)
from pyansys.cell_quality import quality
from pyansys.common import read_binary
from pyansys.convert import convert_script
from pyansys.launcher import launch_mapdl, change_default_ansys_path
from pyansys.misc import Report, _configure_pyvista, _check_has_ansys
from pyansys.examples.downloads import *

# optional www.pyansys.com feature
if 'ANSJUPHUB_VER' in os.environ:
    try:
        from ansys.mapdl import launch_mapdl
    except ImportError:
        raise RuntimeError("you're not supposed to be here")


_HAS_ANSYS = _check_has_ansys()

# Setup data directory
try:
    USER_DATA_PATH = appdirs.user_data_dir('pyansys')
    if not os.path.exists(USER_DATA_PATH):  # pragma: no cover
        os.makedirs(USER_DATA_PATH)

    EXAMPLES_PATH = os.path.join(USER_DATA_PATH, 'examples')
    if not os.path.exists(EXAMPLES_PATH):  # pragma: no cover
        os.makedirs(EXAMPLES_PATH)

except:  # pragma: no cover
    pass

# set pyvista defaults
_configure_pyvista()



"""Various header keys for the result file"""

geometry_header_keys = ['__unused',  # position not used
                        'maxety',
                        'maxrl',
                        'nnod',
                        'nelm',
                        'maxcsy',
                        'ptrETY',
                        'ptrREL',
                        'ptrLOC',
                        'ptrCSY',
                        'ptrEID',
                        'maxsec',
                        'secsiz',
                        'nummat',
                        'matsiz',
                        'ptrMAS',
                        'csysiz',
                        'elmsiz',
                        'etysiz',
                        'rlsiz',
                        'ptrETYl',
                        'ptrETYh',
                        'ptrRELl',
                        'ptrRELh',
                        'ptrCSYl',
                        'ptrCSYh',
                        'ptrLOCl',
                        'ptrLOCh',
                        'ptrEIDl',
                        'ptrEIDh',
                        'ptrMASl',
                        'ptrMASh',
                        'ptrSECl',
                        'ptrSECh',
                        'ptrMATl',
                        'ptrMATh',
                        'ptrCNTl',
                        'ptrCNTh',
                        'ptrNODl',
                        'ptrNODh',
                        'ptrELMl',
                        'ptrELMh',
                        'Glblenb',
                        'ptrGNODl',
                        'ptrGNODh',
                        'maxn',
                        'NodesUpd',
                        'lenbac',
                        'maxcomp',
                        'compsiz',
                        'ptrCOMPl',
                        'ptrCOMPh']

# FROM fdresu.inc
# maxety - the maximum element type reference number in the model
# maxrl  - the maximum real constant reference number in the model
# nnod   - the number of defined nodes in the model
# nelm   - the number of defined elements in the model
# maxcsy - the maximum coordinate system reference number in the model
# ptrETY - pointer to the element type index table
# ptrREL - pointer to the real constant index table
# ptrLOC - pointer to the nodal point locations
# ptrCSY - pointer to the local coordinate system index table
# ptrEID - pointer to the element index table
# maxsec - the maximum section reference number in the model
# secsiz - the maximum size that any section record may have
# nummat - the number of materials in the model
# matsiz - the maximum size that any material property or table may have
# ptrMAS - pointer to the diagonal mass matrix
# csysiz - the number of items describing a local coordinate system (usually 24)
# elmsiz - the maximum number of nodes that a defined element may have
# etysiz - the number of items describing an element type(=IELCSZ from echprm.inc)
# rlsiz  - the maximum number of items defining a real constant (0, if no real constants are defined)
# ETYl,h - 64 bit pointer to element type data
# RELl,h - 64 bit pointer to real constant data
# CSYl,h - 64 bit pointer to coordinate system data
# LOCl,h - 64 bit pointer to nodal locations
# EIDl,h - 64 bit pointer to element data
# SECl,h - 64 bit pointer to section data
# MATl,h - 64 bit pointer to material data
# CNTl,h - 64 bit pointer to element centroids
# NODl,h - 64 bit pointer to nodal equivalence table
# ELMl,h - 64 bit pointer to element equivalence table
# lbnnod - global number of nodes actually used in the solution phase (== nnod unless using Distributed Ansys)
# GNODl,h- 64 bit pointer to the global nodal equivalence table (only used with Distributed ANSYS and when the mesh does not change during solution)
# maxn   - maximum node number of the model


element_index_table_info = {
    'EMS': 'Misc. data',
    'ENF': 'Nodal forces',
    'ENS': 'Nodal stresses',
    'ENG': 'Volume and energies',
    'EGR': 'Nodal gradients',
    'EEL': 'Elastic strains',
    'EPL': 'Plastic strains',
    'ECR': 'Creep strains',
    'ETH': 'Thermal strains',
    'EUL': 'Euler angles',
    'EFX': 'Nodal fluxes',
    'ELF': 'Local forces',
    'EMN': 'Misc. non-sum values',
    'ECD': 'Element current densities',
    'ENL': 'Nodal nonlinear data',
    'EHC': 'Calculated heat generations',
    'EPT': 'Element temperatures',
    'ESF': 'Element surface stresses',
    'EDI': 'Diffusion strains',
    'ETB': 'Etable items',
    'ECT': 'Contact data',
    'EXY': 'Integration point locations',
    'EBA': 'Back stresses',
    'ESV': 'State variables',
    'MNL': 'Material nonlinear record'
}

solution_data_header_keys = ['pv3num', 'nelm', 'nnod', 'mask', 'itime',
                             'iter', 'ncumit', 'nrf', 'cs_LSC', 'nmast',
                             'ptrNSL', 'ptrESL', 'ptrRF', 'ptrMST',
                             'ptrBC', 'rxtrap', 'mode', 'isym', 'kcmplx',
                             'numdof', 'DOFS', 'DOFS', 'DOFS', 'DOFS',
                             'DOFS', 'DOFS', 'DOFS', 'DOFS', 'DOFS',
                             'DOFS', 'DOFS', 'DOFS', 'DOFS', 'DOFS',
                             'DOFS', 'DOFS', 'DOFS', 'DOFS', 'DOFS',
                             'DOFS', 'DOFS', 'DOFS', 'DOFS', 'DOFS',
                             'DOFS', 'DOFS', 'DOFS', 'DOFS', 'DOFS',
                             'DOFS', 'title', 'title', 'title', 'title',
                             'title', 'title', 'title', 'title', 'title',
                             'title', 'title', 'title', 'title', 'title',
                             'title', 'title', 'title', 'title', 'title',
                             'title', 'stitle', 'stitle', 'stitle',
                             'stitle', 'stitle', 'stitle', 'stitle',
                             'stitle', 'stitle', 'stitle', 'stitle',
                             'stitle', 'stitle', 'stitle', 'stitle',
                             'stitle', 'stitle', 'stitle', 'stitle',
                             'stitle', 'dbmtim', 'dbmdat', 'dbfncl',
                             'soltim', 'soldat', 'ptrOND', 'ptrOEL',
                             'nfldof', 'ptrEXA', 'ptrEXT', 'ptrEXAl',
                             'ptrEXAh', 'ptrEXTl', 'ptrEXTh', 'ptrNSLl',
                             'ptrNSLh', 'ptrRFl', 'ptrRFh', 'ptrMSTl',
                             'ptrMSTh', 'ptrBCl', 'ptrBCh', 'ptrTRFl',
                             'ptrTRFh', 'ptrONDl', 'ptrONDh', 'ptrOELl',
                             'ptrOELh', 'ptrESLl', 'ptrESLh', 'ptrOSLl',
                             'ptrOSLh', 'sizeDEAD', 'ptrDEADl', 'ptrDEADh',
                             'PrinKey', 'numvdof', 'numadof', '0', '0',
                             'ptrVSLl', 'ptrVSLh', 'ptrASLl', 'ptrASLh', '0',
                             '0', '0', '0', 'numRotCmp', '0',
                             'ptrRCMl', 'ptrRCMh', 'nNodStr', '0', 'ptrNDSTRl',
                             'ptrNDSTRh', 'AvailData', 'geomID', 'ptrGEOl', 'ptrGEOh']

solution_header_keys_dp = ['timfrq', 'lfacto', 'lfactn', 'cptime',
                           'tref', 'tunif', 'tbulk', 'volbase',
                           'tstep', '__unused', 'accel_x', 'accel_y',
                           'accel_z', 'omega_v_x', 'omega_v_y',
                           'omega_v_z', 'omega_a_x', 'omega_a_y',
                           'omega_a_z', 'omegacg_v_x', 'omegacg_v_y',
                           'omegacg_v_z', 'omegacg_a_x',
                           'omegacg_a_y', 'omegacg_a_z', 'cgcent',
                           'cgcent', 'cgcent', 'fatjack', 'fatjack',
                           'dval1', 'pCnvVal']

result_header_keys = ['fun12', 'maxn', 'nnod', 'resmax', 'numdof',
                      'maxe', 'nelm', 'kan', 'nsets', 'ptrend',
                      'ptrDSIl', 'ptrTIMl', 'ptrLSPl', 'ptrELMl',
                      'ptrNODl', 'ptrGEOl', 'ptrCYCl', 'CMSflg',
                      'csEls', 'units', 'nSector', 'csCord',
                      'ptrEnd8', 'ptrEnd8', 'fsiflag', 'pmeth',
                      'noffst', 'eoffst', 'nTrans', 'ptrTRANl',
                      'PrecKey', 'csNds', 'cpxrst', 'extopt',
                      'nlgeom', 'AvailData', 'mmass', 'kPerturb',
                      'XfemKey', 'rstsprs', 'ptrDSIh', 'ptrTIMh',
                      'ptrLSPh', 'ptrCYCh', 'ptrELMh', 'ptrNODh',
                      'ptrGEOh', 'ptrTRANh', 'Glbnnod', 'ptrGNODl',
                      'ptrGNODh', 'qrDmpKy', 'MSUPkey', 'PSDkey',
                      'cycMSUPkey', 'XfemCrkPropTech']

boundary_condition_index_table = [
    'numdis',  # number of nodal constraints
    'ptrDIX',  # pointer to the table of nodes having nodal constraints
    'ptrDIS',  # pointer to nodal constraint values
    'numfor',  # number of nodal input force loadings
    'ptrFIX',  # pointer to the table of nodes having nodal forces
    'ptrFOR',  # pointer to nodal force values
    'format'  # key (0 or 1) denoting which format is used for DIX/FIX data (see below)
]
# if format == 0 --> Nodal constraint DOF.
# This index is calculated as N*32+DOF, where N is the node number and
# DOF is the DOF reference number.  Values are in the same order as
# the DOF number reference table.
# if format == 1 --> Nodal constraint node
# numbers.

# Degrees of freedom per node DOF reference numbers are
DOF_REF = {
    1: 'UX',
    2: 'UY',
    3: 'UZ',
    4: 'ROTX',
    5: 'ROTY',
    6: 'ROTZ',
    7: 'AX',
    8: 'AY',
    9: 'AZ',
    10: 'VX',
    11: 'VY',
    12: 'VZ',
    16: 'WARP',
    17: 'CONC',
    18: 'HDSP',
    19: 'PRES',
    20: 'TEMP',
    21: 'VOLT',
    22: 'MAG',
    23: 'ENKE',
    24: 'ENDS',
    25: 'EMF',
    26: 'CURR',
    27: 'SP01',
    28: 'SP02',
    29: 'SP03',
    30: 'SP04',
    31: 'SP05',
    32: 'SP06',
    33: 'TBOT',
    34: 'TE2',
    35: 'TE3',
    36: 'TE4',
    37: 'TE5',
    38: 'TE6',
    39: 'TE7',
    40: 'TE8',
    41: 'TE9',
    42: 'TE10',
    43: 'TE11',
    44: 'TE12',
    45: 'TE13',
    46: 'TE14',
    47: 'TE15',
    48: 'TE16',
    49: 'TE17',
    50: 'TE18',
    51: 'TE19',
    52: 'TE20',
    53: 'TE21',
    54: 'TE22',
    55: 'TE23',
    56: 'TE24',
    57: 'TE25',
    58: 'TE26',
    59: 'TE27',
    60: 'TE28',
    61: 'TE29',
    62: 'TE30',
    63: 'TE31',
    64: 'TTOP'
}



"""Handle result files from a distributed MAPDL analysis"""
from inspect import currentframe
import glob
import os
from functools import wraps

import pyvista as pv
import numpy as np
from vtk import vtkAppendFilter

from pyansys.misc import is_float, vtk_cell_info
from pyansys.mesh import Mesh
from pyansys.rst import Result, ELEMENT_INDEX_TABLE_KEYS
from pyansys.errors import NoDistributedFiles
from pyansys._binary_reader import (read_nodal_values_dist,
                                    populate_surface_element_result)
from pyansys._rst_keys import element_index_table_info


def find_dis_files(main_file):
    """Find the individual distributed result files given a main result file"""
    basename = os.path.basename(main_file)
    jobname = basename[:basename.rfind('0')]
    ext = basename.split('.')[1]
    dirname = os.path.dirname(main_file)

    if dirname is None:
        dirname = os.getcwd()

    # get all files matching the jobname
    filenames = {}
    for filename in glob.glob('%s/%s*.%s' % (dirname, jobname, ext)):
        index = os.path.basename(filename).replace(jobname, '')[:-4]
        if is_float(index):
            filenames[int(index)] = filename

    if not filenames:
        raise NoDistributedFiles

    if max(filenames.keys()) + 1 != len(filenames):
        raise FileNotFoundError('Unable to find all the result files of a '
                                'distributed result')

    return filenames


class DistributedResult(Result):
    """Distributed result file

    Parameters
    ----------
    main_file : str
        Path of main result file

    """

    def __init__(self, main_file):
        """Initialize from a series of distributed files"""
        # find remainder of distributed results
        filenames = find_dis_files(main_file)

        # load initial result
        super().__init__(main_file, read_mesh=False)
        self._results = [Result(main_file)]

        # Global number of nodes must not equal the number of nodes in this file
        if not self._main_result._is_distributed:
            raise RuntimeError('Result file is not part of a distributed result')

        if not self._main_result._is_main:  # pragma: no cover
            raise RuntimeError('DistributedResult must be created from the main '
                               'result file')

        # load and verify
        ptr = self._main_result._resultheader['ptrGNOD']
        gl_nnum = self._main_result.read_record(ptr)

        mask = np.in1d(gl_nnum, self._main_result.mesh.nnum, assume_unique=True)
        for index in range(1, len(filenames)):
            result = Result(filenames[index])
            new_mask = np.in1d(gl_nnum, result.mesh.nnum, assume_unique=True)
            if not new_mask.any():  # pragma: no cover
                raise RuntimeError('File %s not part of the distributed result'
                                   % filenames[index])
            mask += new_mask
            self._results.append(result)

        if not np.all(mask):
            filenames = '\n'.join([result.filename for result in self._results])
            raise FileNotFoundError('Total nodes loaded from the individual result '
                                    'files does not match the number in the global '
                                    'index.  \n\nResult files found include:\n\n%s' %
                                    filenames)

        # assemble the global mesh
        vtkappend = vtkAppendFilter()
        vtkappend.MergePointsOn()
        st = 0
        for i, result in enumerate(self._results):
            # map the merged node numbering with the individual mappings
            result.grid['_dist_idx'] = np.arange(st, st + result.grid.n_points)
            _result_idx = np.empty(result.grid.n_cells, np.int32)
            _result_idx[:] = i
            result.grid.cell_arrays['_result_idx'] = _result_idx
            st += result.grid.n_points
            vtkappend.AddInputData(result.grid)

        self._total_sol_nodes = st  # total nodes in all result files

        vtkappend.Update()
        grid = pv.wrap(vtkappend.GetOutput())
        # nodes are not sorted here

        # Map nodes from individual results to the global mesh
        self._glb_idx = grid['_dist_idx']

        # start of each section of the grid
        # elem_split_ind = np.diff(self.grid['_result_idx']).nonzero()[0]
        # self._elem_split = np.hstack(([0], elem_split_ind))

        elem = np.hstack(result.mesh._elem for result in self._results)
        glb_elem_off = []
        for result in self._results:
            elem_off = result.mesh._elem_off
            if len(glb_elem_off):
                off = glb_elem_off[-1][-1]
                glb_elem_off.append(elem_off[1:] + off)
            else:
                glb_elem_off.append(elem_off)

        # TODO: Add node and element components
        # resort node numbers
        self._dist_sort_idx = np.argsort(grid['ansys_node_num'])
        self._sorted_nnum = grid['ansys_node_num'][self._dist_sort_idx]
        nodes = grid.points[self._dist_sort_idx]

        glb_elem_off = np.hstack(glb_elem_off)
        self._mesh = Mesh(self._sorted_nnum, nodes, elem, glb_elem_off,
                          self._main_result.mesh.ekey)
                          # node_comps=ncomp, elem_comps=ecomp)

        self.quadgrid = self._mesh._parse_vtk(fix_midside=False)
        self.grid = self.quadgrid.linear_copy()

        # self._neqv = self._resultheader['neqv']  # may not need this
        self._eeqv = self.grid.cell_arrays['ansys_elem_num']

    @property
    def _main_result(self):
        """Main result instance"""
        return self._results[0]

    @property
    @wraps(Result.mesh)
    def mesh(self):
        """Assembled global mesh is an assembly of the individual result file meshes"""
        return self._mesh

    @wraps(Result.nodal_solution)
    def nodal_solution(self, *args, **kwargs):
        return self._dis_solution(currentframe().f_code.co_name, *args, **kwargs)

    def _dis_solution(self, func_name, *args, **kwargs):
        """Get the distributed solution for a given function"""
        glb_nnum = []
        glb_sol = []
        for result in self._results:
            func = getattr(result, func_name)
            rst_nnum, rst_sol = func(*args, **kwargs)
            glb_nnum.append(rst_nnum)
            glb_sol.append(rst_sol)

        # resort and organize stacked solutions
        glb_sol = np.vstack(glb_sol)
        if glb_sol.shape[0] == self._total_sol_nodes:
            # not all values from the global solution are unique and
            # we need to pare this down (due to merging to meshes)
            glb_sol = glb_sol[self._glb_idx]
            return self._sorted_nnum, glb_sol[self._dist_sort_idx]

            # return self.mesh.nnum, glb_sol
        else:  # limited subset of solution
            # must remap due to how the subset of nodes relate to the
            # global mesh solution
            nnum = np.hstack(glb_nnum)
            u_nnum, idx = np.unique(nnum, return_index=True)
            return u_nnum, glb_sol[idx]

    def _nodal_result(self, rnum, result_type, **kwargs):
        """Load generic nodal result

        Parameters
        ----------
        rnum : int
            Result number.

        result_type : int
            EMS: misc. data
            ENF: nodal forces
            ENS: nodal stresses
            ENG: volume and energies
            EGR: nodal gradients
            EEL: elastic strains
            EPL: plastic strains
            ECR: creep strains
            ETH: thermal strains
            EUL: euler angles
            EFX: nodal fluxes
            ELF: local forces
            EMN: misc. non-sum values
            ECD: element current densities
            ENL: nodal nonlinear data
            EHC: calculated heat
            EPT: element temperatures
            ESF: element surface stresses
            EDI: diffusion strains
            ETB: ETABLE items(post1 only
            ECT: contact data
            EXY: integration point locations
            EBA: back stresses
            ESV: state variables
            MNL: material nonlinear record

        Returns
        -------
        nnum : np.ndarray
            ANSYS node numbers

        result : np.ndarray
            Array of result data
        """
        # check result exists
        if not self.available_results[result_type]:
            raise ValueError('Result %s is not available in this result file'
                             % result_type)

        # element header
        rnum = self.parse_step_substep(rnum)

        result_type = result_type.upper()
        nitem = self._result_nitem(rnum, result_type)
        result_index = ELEMENT_INDEX_TABLE_KEYS.index(result_type)

        # Element types for nodal averaging from the global mesh
        n_points = self.grid.n_points
        cells, offset = vtk_cell_info(self.grid)
        data = np.zeros((n_points, nitem), np.float64)
        ncount = np.zeros(n_points, np.int32)

        c = 0  # global cell index counter
        for result in self._results:
            ele_ind_table, nodstr, etype, ptr_off = result._element_solution_header(rnum)
            # we return c here since it is copied to C, not passed by reference
            c = read_nodal_values_dist(result.filename,
                                       self.grid.celltypes,
                                       ele_ind_table,
                                       offset,
                                       cells,
                                       nitem,
                                       n_points,
                                       nodstr,
                                       etype,
                                       self._mesh.etype,
                                       result_index,
                                       ptr_off,
                                       ncount,
                                       data,
                                       c)

        if result_type == 'ENS' and nitem != 6:
            data = data[:, :6]

        if not np.any(ncount):
            raise ValueError('Result file contains no %s records for result %d' %
                             (element_index_table_info[result_type.upper()], rnum))

        # average across nodes
        data /= ncount.reshape(-1, 1)
        return self.grid.point_arrays['ansys_node_num'], data

    @wraps(Result.element_solution_data)
    def element_solution_data(self, *args, **kwargs):
        """Accumulate the element solution individual results from each result"""
        sort = kwargs.get('sort', True)
        glb_element_data = []
        for result in self._results:
            element_data = result.element_solution_data(*args, is_dist_rst=True,
                                                        **kwargs)
            glb_element_data.extend(element_data)

        # Assemble and sort (args[0] is rnum)
        _, nodstr, etype, _ = self._element_solution_header(args[0])

        enum = self._eeqv
        enode = []
        nnode = nodstr[etype]
        if sort:
            sidx = np.argsort(enum)
            enum = enum[sidx]
            glb_element_data = [glb_element_data[i] for i in sidx]

            for i in sidx:
                enode.append(self._mesh.elem[i][10:10+nnode[i]])

        else:
            for i in range(enum.size):
                enode.append(self._mesh.elem[i][10:10+nnode[i]])

        return enum, glb_element_data, enode

    @wraps(Result.element_stress)
    def element_stress(self, *args, **kwargs):
        """Accumulate the element solution individual results from each result"""
        sort = kwargs.get('sort', True)
        glb_stress = []
        for result in self._results:
            element_data = result.element_stress(*args, is_dist_rst=True, **kwargs)
            glb_stress.extend(element_data)

        # Assemble and sort (args[0] is rnum)
        _, nodstr, etype, _ = self._element_solution_header(args[0])

        enum = self._eeqv
        enode = []
        nnode = nodstr[etype]

        if sort:
            sidx = np.argsort(enum)
            enum = enum[sidx]
            glb_element_data = [glb_stress[i] for i in sidx]

            for i in sidx:
                enode.append(self._mesh.elem[i][10:10+nnode[i]])

        else:
            for i in range(enum.size):
                enode.append(self._mesh.elem[i][10:10+nnode[i]])

        return enum, glb_element_data, enode

    def plot_element_result(self, rnum, result_type, item_index,
                            in_element_coord_sys=False, **kwargs):
        """Plot an element result.

        Parameters
        ----------
        rnum : int
            Result number.

        result_type : str
            Element data type to retreive.

            - EMS: misc. data
            - ENF: nodal forces
            - ENS: nodal stresses
            - ENG: volume and energies
            - EGR: nodal gradients
            - EEL: elastic strains
            - EPL: plastic strains
            - ECR: creep strains
            - ETH: thermal strains
            - EUL: euler angles
            - EFX: nodal fluxes
            - ELF: local forces
            - EMN: misc. non-sum values
            - ECD: element current densities
            - ENL: nodal nonlinear data
            - EHC: calculated heat generations
            - EPT: element temperatures
            - ESF: element surface stresses
            - EDI: diffusion strains
            - ETB: ETABLE items
            - ECT: contact data
            - EXY: integration point locations
            - EBA: back stresses
            - ESV: state variables
            - MNL: material nonlinear record

        item_index : int
            Index of the data item for each node within the element.

        in_element_coord_sys : bool, optional
            Returns the results in the element coordinate system.
            Default False and will return the results in the global
            coordinate system.

        Returns
        -------
        nnum : np.ndarray
            ANSYS node numbers

        result : np.ndarray
            Array of result data

        Examples
        --------
        # >>> rst.plot_element_result(
        """
        # check result exists
        result_type = result_type.upper()
        if not self.available_results[result_type]:
            raise ValueError('Result %s is not available in this result file'
                             % result_type)

        if result_type not in ELEMENT_INDEX_TABLE_KEYS:
            raise ValueError('Result "%s" is not an element result' % result_type)

        bsurfs = []
        for result in self._results:
            bsurf = result._extract_surface_element_result(rnum,
                                                           result_type,
                                                           item_index,
                                                           in_element_coord_sys)
            bsurfs.append(bsurf)

        desc = self.available_results.description[result_type].capitalize()
        kwargs.setdefault('stitle', desc)
        return pv.plot(bsurfs, scalars='_scalars', **kwargs)



"""Functions to download sampe datasets from the VTK data repository
"""
import shutil
import os
import urllib.request
import zipfile

import pyansys


def get_ext(filename):
    """Extract the extension of the filename"""
    ext = os.path.splitext(filename)[1].lower()
    return ext


def delete_downloads():
    """Delete all downloaded examples to free space or update the files"""
    shutil.rmtree(pyansys.EXAMPLES_PATH)
    os.makedirs(pyansys.EXAMPLES_PATH)
    return True


def _decompress(filename):
    zip_ref = zipfile.ZipFile(filename, 'r')
    zip_ref.extractall(pyansys.EXAMPLES_PATH)
    return zip_ref.close()


def _get_vtk_file_url(filename):
    return 'https://github.com/akaszynski/pyansys-data/raw/master/data/{}'.format(filename)


def _retrieve_file(url, filename):
    # First check if file has already been downloaded
    local_path = os.path.join(pyansys.EXAMPLES_PATH, os.path.basename(filename))
    local_path_no_zip = local_path.replace('.zip', '')
    if os.path.isfile(local_path_no_zip) or os.path.isdir(local_path_no_zip):
        return local_path_no_zip, None

    # grab the correct url retriever
    urlretrieve = urllib.request.urlretrieve

    # Perfrom download
    saved_file, resp = urlretrieve(url)
    shutil.move(saved_file, local_path)
    if get_ext(local_path) in ['.zip']:
        _decompress(local_path)
        local_path = local_path[:-4]
    return local_path, resp


def download_file(filename):
    return _download_file(filename)


def _download_file(filename):
    url = _get_vtk_file_url(filename)
    return _retrieve_file(url, filename)


def _download_and_read(filename):
    saved_file, _ = _download_file(filename)
    if saved_file[-3:] == 'cdb':
        return pyansys.Archive(saved_file)
    else:
        return pyansys.read_binary(saved_file)


###############################################################################
def download_verification_result(index):
    """Download a verification manual result file"""
    return _download_and_read('vm%d.rst' % index)


def download_shaft_modal():
    """Download modal analysis of a rotor shaft"""
    return _download_and_read('shaft_modal.rst')


def download_sector_modal():
    """Download modal analysis of a cyclic turbine sector"""
    return _download_and_read('sector_modal.rst')


def download_pontoon():
    """Download modal analysis of a cyclic turbine sector"""
    return _download_and_read('pontoon.rst')


def _download_solid239_240():
    """SOLID239 and SOLID240 file for unit testing"""
    return _download_and_read('solid239_240.rth')


def _download_plane238():
    """PLANE result for unit testing"""
    return _download_and_read('plane238.rth')


def _download_shell181():
    """SHELL181 result for unit testing"""
    return _download_and_read('shell181.rst')


def download_corner_pipe():
    """Corner pipe result for unit testing and basic demo"""
    return _download_and_read('cyc_stress.rst')


def download_academic_rotor_result():
    """Cyclic academic rotor result file"""
    return _download_and_read('academic_rotor.rst')


def download_academic_rotor_archive():
    """Cyclic academic rotor archive file"""
    return _download_and_read('academic_rotor.cdb')


def download_academic_rotor_4blade_result():
    """Cyclic academic rotor result file"""
    return _download_and_read('academic_rotor_4_blade.rst')
